import re
from abc import ABC, abstractmethod
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from poke_env.battle.effect import Effect
from poke_env.battle.field import Field
from poke_env.battle.move import Move
from poke_env.battle.pokemon import Pokemon
from poke_env.battle.pokemon_type import PokemonType
from poke_env.battle.side_condition import STACKABLE_CONDITIONS, SideCondition
from poke_env.battle.weather import Weather
from poke_env.data import GenData, to_id_str
from poke_env.data.replay_template import REPLAY_TEMPLATE
from poke_env.teambuilder.teambuilder_pokemon import TeambuilderPokemon


class AbstractBattle(ABC):
    MESSAGES_TO_IGNORE = {
        "-anim",
        "-block",
        "-burst",
        "-center",
        "-combine",
        "-crit",
        "-fail",
        "-fieldactivate",
        "-hint",
        "-hitcount",
        "-miss",
        "-notarget",
        "-nothing",
        "-ohko",
        "-resisted",
        "-supereffective",
        "-waiting",
        "-zbroken",
        "J",
        "L",
        "askreg",
        "badge",
        "c",
        "chat",
        "crit",
        "debug",
        "deinit",
        "gametype",
        "hidelines",
        "html",
        "immune",
        "inactiveoff",
        "init",
        "j",
        "join",
        "l",
        "leave",
        "n",
        "name",
        "rated",
        "resisted",
        "sentchoice",
        "split",
        "supereffective",
        "teampreview",
        "upkeep",
        "uhtml",
        "zbroken",
        "",
    }

    __slots__ = (
        "_anybody_inactive",
        "_available_moves",
        "_available_switches",
        "_battle_tag",
        "_can_dynamax",
        "_can_mega_evolve",
        "_can_tera",
        "_can_z_move",
        "_commanding",
        "_dynamax_turn",
        "_fields",
        "_finished",
        "_force_switch",
        "_format",
        "_gen",
        "in_team_preview",
        "_last_request",
        "_max_team_size",
        "_maybe_trapped",
        "_opponent_dynamax_turn",
        "_opponent_rating",
        "_opponent_side_conditions",
        "_opponent_team",
        "_opponent_used_dynamax",
        "_opponent_used_mega_evolve",
        "_opponent_used_tera",
        "_opponent_used_z_move",
        "_opponent_username",
        "_player_role",
        "_player_username",
        "_players",
        "_rating",
        "_reconnected",
        "_replay_data",
        "rules",
        "_reviving",
        "_save_replays",
        "_side_conditions",
        "_team_size",
        "_team",
        "_teambuilder_team",
        "_teampreview_team",
        "_teampreview_opponent_team",
        "_teampreview",
        "_trapped",
        "_turn",
        "_used_dynamax",
        "_used_mega_evolve",
        "_used_tera",
        "_used_z_move",
        "_wait",
        "_weather",
        "_won",
        "logger",
    )

    def __init__(
        self,
        battle_tag: str,
        username: str,
        logger: Logger,
        save_replays: Union[str, bool],
        gen: int,
    ):
        # Utils attributes
        self._battle_tag: str = battle_tag
        self._gen: int = gen
        self._format: Optional[str] = None
        self._max_team_size: Optional[int] = None
        self._opponent_username: Optional[str] = None
        self._player_role: Optional[str] = None
        self._player_username: str = username
        self._players: List[Dict[str, str]] = []
        self._replay_data: List[List[str]] = []
        self._save_replays: Union[str, bool] = save_replays
        self._team_size: Dict[str, int] = {}
        self._teampreview: bool = False
        self._teampreview_team: List[Pokemon] = []
        self._teampreview_opponent_team: List[Pokemon] = []
        self._anybody_inactive: bool = False
        self._reconnected: bool = True
        self.logger: Optional[Logger] = logger

        # Turn choice attributes
        self.in_team_preview: bool = False
        self._wait: bool = False

        # Battle state attributes
        self._dynamax_turn: Optional[int] = None
        self._finished: bool = False
        self._last_request: Dict[str, Any] = {}
        self.rules: List[str] = []
        self._turn: int = 0
        self._opponent_dynamax_turn: Optional[int] = None
        self._opponent_rating: Optional[int] = None
        self._rating: Optional[int] = None
        self._won: Optional[bool] = None

        # In game battle state attributes
        self._weather: Dict[Weather, int] = {}
        self._fields: Dict[Field, int] = {}  # set()
        self._opponent_side_conditions: Dict[SideCondition, int] = {}  # set()
        self._side_conditions: Dict[SideCondition, int] = {}  # set()
        self._reviving: bool = False
        self._commanding: bool = False
        self._opponent_used_mega_evolve = False
        self._opponent_used_z_move = False
        self._opponent_used_dynamax = False
        self._opponent_used_tera = False
        self._used_mega_evolve = False
        self._used_z_move = False
        self._used_dynamax = False
        self._used_tera = False

        # Pokemon attributes
        self._team: Dict[str, Pokemon] = {}
        self._teambuilder_team: list[TeambuilderPokemon] | None = None
        self._opponent_team: Dict[str, Pokemon] = {}

    def get_pokemon(
        self,
        identifier: str,
        force_self_team: bool = False,
        details: str = "",
        request: Optional[Dict[str, Any]] = None,
    ) -> Pokemon:
        """Returns the Pokemon object corresponding to given identifier. Can force to
        return object from the player's team if force_self_team is True. If the Pokemon
        object does not exist, it will be created. Details can be given, which is
        necessary to initialize alternate forms (eg. alolan pokemons) properly.

        :param identifier: The identifier to use to retrieve the pokemon.
        :type identifier: str
        :param force_self_team: Wheter to force returning a Pokemon from the player's
            team. Defaults to False.
        :type force_self_team: bool
        :param details: Detailled information about the pokemon. Defaults to ''.
        :type details: str, defaults to ''
        :param request: Detailled information about the pokemon from a request.
            Defaults to None.
        :type request: Dict, optional, defaults to None
        :return: The corresponding pokemon object.
        :rtype: Pokemon
        :raises ValueError: If the team has too many pokemons, as determined by the
            teamsize component of battle initialisation.
        """
        if identifier[3] != " ":
            identifier = identifier[:2] + identifier[3:]

        if identifier in self._team:
            return self._team[identifier]
        elif identifier in self._opponent_team:
            return self._opponent_team[identifier]

        player_role = identifier[:2]
        name = identifier[3:].strip()
        team = (
            self._team
            if player_role == self.player_role or force_self_team
            else self._opponent_team
        )

        # if the pokemon has a nickname, this ensures we recognize it
        name_det = details.split(", ")[0]
        matches = [i for i, p in enumerate(team.values()) if p.identifies_as(name_det)]
        assert len(matches) < 2
        if identifier not in team and matches:
            i = matches[0]
            items = list(team.items())
            items[i] = (identifier, items[i][1])
            items[i][1]._name = identifier[4:]
            if player_role == self.player_role or force_self_team:
                self._team = dict(items)
            else:
                self._opponent_team = dict(items)
        team = (
            self._team
            if player_role == self.player_role or force_self_team
            else self._opponent_team
        )
        if identifier in team:
            return team[identifier]

        if self._team_size and len(team) >= self._team_size[player_role]:
            raise ValueError(
                "%s's team already has %d pokemons: cannot add %s to %s"
                % (
                    player_role,
                    self._team_size[player_role],
                    identifier,
                    ", ".join(team.keys()),
                )
            )

        if request:
            team[identifier] = Pokemon(request_pokemon=request, name=name, gen=self.gen)
        elif details:
            team[identifier] = Pokemon(details=details, name=name, gen=self.gen)
        else:
            species = identifier[4:]
            team[identifier] = Pokemon(species=species, name=name, gen=self.gen)

        return team[identifier]

    @abstractmethod
    def clear_all_boosts(self):
        pass

    def _check_damage_message_for_item(self, split_message: List[str]):
        # Catches when a side takes damage from the opponent's item
        # The item belongs to the side not taking damage
        # Example:
        #   |-damage|p2a: Archeops|88/100|[from] item: Rocky Helmet|[of] p1a: Ferrothorn
        if (
            len(split_message) == 6
            and split_message[4].startswith("[from] item:")
            and split_message[5].startswith("[of]")
        ):
            item = split_message[4].split("item:")[-1]
            pkmn = split_message[5].split("[of]")[-1].strip()
            self.get_pokemon(pkmn).item = to_id_str(item)

        # Catches when a side takes damage from it's own item
        # The item belongs to the same side taking damage
        # Example:
        #   |-damage|p2a: Pikachu|90/100|[from] item: Life Orb
        elif len(split_message) == 5 and split_message[4].startswith("[from] item:"):
            item = split_message[4].split("item:")[-1]
            pkmn = split_message[2]
            self.get_pokemon(pkmn).item = to_id_str(item)

    def _check_damage_message_for_ability(self, split_message: List[str]):
        # Catches when a side takes damage from the opponent's ability
        # the item is from the side not taking damage
        # Example:
        #   |-damage|p2a: Archeops|88/100|[from] ability: Iron Barbs|[of] p1a: Ferrothorn
        if (
            len(split_message) == 6
            and split_message[4].startswith("[from] ability:")
            and split_message[5].startswith("[of]")
        ):
            ability = split_message[4].split("ability:")[-1]
            pkmn = split_message[5].split("[of]")[-1].strip()
            self.get_pokemon(pkmn).ability = ability

    def _check_heal_message_for_item(self, split_message: List[str]):
        # Catches when a side heals from it's own item
        # the check for item is not None is necessary because the PS simulator will
        #  show the heal message AFTER a berry has already been consumed
        # Examples:
        #  |-heal|p2a: Quagsire|100/100|[from] item: Leftovers
        #  |-heal|p2a: Quagsire|100/100|[from] item: Sitrus Berry
        if len(split_message) == 5 and split_message[4].startswith("[from] item:"):
            pkmn = split_message[2]
            item = to_id_str(split_message[4].split("item:")[-1])
            pkmn_object = self.get_pokemon(pkmn)
            if (
                pkmn_object.item is not None
                # don't assign an item that was just consumed
                and "berry" not in item
                and "herb" not in item
            ):
                pkmn_object.item = item

    def _check_heal_message_for_ability(self, split_message: List[str]):
        # Catches when a side heals from it's own ability
        # the "of" component sent by the PS server is a bit misleading
        #   it implies the ability is from the opposite side
        # Example:
        #   |-heal|p2a: Quagsire|100/100|[from] ability: Water Absorb|[of] p1a: Genesect
        #   |-heal|p2b: Excadrill|100/100|from] ability: Hospitality|[of] p2a: Sinistcha
        if len(split_message) == 6 and split_message[4].startswith("[from] ability:"):
            ability = to_id_str(split_message[4].split("ability:")[-1])
            if ability == "hospitality":
                pkmn = split_message[5].replace("[of] ", "").strip()
                self.get_pokemon(pkmn).ability = ability
            else:
                pkmn = split_message[2]
                self.get_pokemon(pkmn).ability = ability

    @abstractmethod
    def end_illusion(self, pokemon_name: str, details: str):
        pass

    def _end_illusion_on(
        self, illusionist: Optional[str], illusioned: Optional[Pokemon], details: str
    ):
        if illusionist is None:
            raise ValueError("Cannot end illusion without an active pokemon.")
        if illusioned is None:
            raise ValueError("Cannot end illusion without an illusioned pokemon.")
        illusionist_mon = self.get_pokemon(illusionist, details=details)

        if illusionist_mon is illusioned:
            return illusionist_mon

        illusionist_mon.switch_in(details=details)
        illusionist_mon.status = illusioned.status
        illusionist_mon.set_hp(f"{illusioned.current_hp}/{illusioned.max_hp}")

        illusioned.was_illusioned(self.fields)

        return illusionist_mon

    def _field_end(self, field_str: str):
        field = Field.from_showdown_message(field_str)
        if field is not Field.UNKNOWN:
            if field is Field.NEUTRALIZING_GAS:
                self._fields.pop(field, 0)
            else:
                self._fields.pop(field)

    def field_start(self, field_str: str):
        field = Field.from_showdown_message(field_str)

        if field.is_terrain:
            self._fields = {
                field: turn
                for field, turn in self._fields.items()
                if not field.is_terrain
            }

        self._fields[field] = self.turn

    @staticmethod
    def _split_message_to_replay_event(split_message: List[str]) -> str:
        return "|".join(split_message)

    @staticmethod
    def _replay_event_type(replay_event: str) -> str:
        if not replay_event.startswith("|"):
            return ""
        replay_event_parts = replay_event.split("|", 2)
        if len(replay_event_parts) <= 1:
            return ""
        return replay_event_parts[1]

    def _has_terminal_replay_result(self) -> bool:
        return any(
            len(split_message) > 1 and split_message[1] in {"win", "tie"}
            for split_message in self._replay_data
        )

    def _build_replay_events(self) -> List[str]:
        replay_events = [
            self._split_message_to_replay_event(split_message)
            for split_message in self._replay_data
        ]

        # Fallback for cases where battle result is known but terminal replay event is
        # missing from tracked protocol messages.
        if self.finished and not any(
            self._replay_event_type(event) in {"win", "tie"} for event in replay_events
        ):
            if self._won is True:
                replay_events.append(f"|win|{self._player_username}")
            elif self._won is False:
                replay_events.append(f"|win|{self._opponent_username or 'Opponent'}")
            else:
                replay_events.append("|tie")

        return replay_events

    def _build_replay_log(self) -> str:
        replay_events = self._build_replay_events()
        replay_log = f">{self.battle_tag}"
        if replay_events:
            replay_log = f"{replay_log}\n" + "\n".join(replay_events)
        return replay_log

    def _build_replay_html(self) -> str:
        formatted_replay = REPLAY_TEMPLATE
        formatted_replay = formatted_replay.replace(
            "{BATTLE_TAG}", f"{self.battle_tag}"
        )
        formatted_replay = formatted_replay.replace(
            "{PLAYER_USERNAME}", f"{self._player_username}"
        )
        formatted_replay = formatted_replay.replace(
            "{OPPONENT_USERNAME}", f"{self._opponent_username or 'Opponent'}"
        )
        formatted_replay = formatted_replay.replace(
            "{REPLAY_LOG}", self._build_replay_log()
        )

        return formatted_replay

    def save_replay(self, file_path: Union[str, Path]) -> Path:
        """
        Writes this battle replay to a file.

        :param file_path: Path where replay html should be written.
        :type file_path: str | pathlib.Path
        :return: The written replay path.
        :rtype: pathlib.Path
        """
        replay_path = Path(file_path)
        replay_path.parent.mkdir(parents=True, exist_ok=True)
        replay_path.write_text(self._build_replay_html(), encoding="utf-8")
        return replay_path

    def _finish_battle(self):
        self._finished = True

        if self._save_replays:
            if self._save_replays is True:
                folder = "replays"
            else:
                folder = str(self._save_replays)

            self.save_replay(
                Path(folder) / f"{self._player_username} - {self.battle_tag}.html"
            )

    @abstractmethod
    def _get_target_mon(
        self, pokemon: str, target_type: str, target_str: str | None
    ) -> Pokemon | None:
        pass

    def is_grounded(self, mon: Pokemon):
        if Field.GRAVITY in self.fields:
            return True
        elif mon.item == "ironball":
            return True
        elif mon.ability == "levitate":
            return False
        elif mon.ability is None and "levitate" in mon.possible_abilities:
            return False
        elif mon.item == "airballoon":
            return False
        elif mon.type_1 == PokemonType.FLYING or mon.type_2 == PokemonType.FLYING:
            return False
        elif Effect.MAGNET_RISE in mon.effects:
            return False
        return True

    def parse_message(self, split_message: List[str]):
        self._replay_data.append(split_message[:])

        # We copy because we directly modify split_message in poke-env; this is to
        # preserve further usage of this event upstream
        event = split_message[:]

        if event[1] in self.MESSAGES_TO_IGNORE:
            return
        elif event[1] in ["drag", "switch"]:
            pokemon, details, hp_status = event[2:5]
            self.switch(pokemon, details, hp_status)
        elif event[1] == "-damage":
            pokemon, hp_status = event[2:4]
            self.get_pokemon(pokemon).damage(hp_status)
            self._check_damage_message_for_item(event)
            self._check_damage_message_for_ability(event)
        elif event[1] == "move":
            pokemon = event[2]
            mon = self.get_pokemon(pokemon)
            use = not mon._dancing
            failed = False
            reveal = not mon._dancing
            overridden_move = None
            spread = False
            mon._dancing = False

            for move_failed_suffix in ["[miss]", "[still]", "[notarget]"]:
                if event[-1] == move_failed_suffix:
                    event = event[:-1]
                    failed = True

            if event[-1] == "[notarget]":
                event = event[:-1]

            while event[-1].startswith("[spread]"):
                spread = True
                event = event[:-1]

            if event[-1] in {
                "[from] lockedmove",
                "[from]lockedmove",
                "[from] Sky Attack",
            }:
                use = False
                reveal = False
                event = event[:-1]

            if event[-1] in {"[from] Pursuit", "[from]Pursuit", "[zeffect]"}:
                event = event[:-1]

            if event[-1] == "[from] Sleep Talk":
                event[-1] = "[from] move: Sleep Talk"

            if event[-1].startswith("[anim]"):
                event = event[:-1]

            if event[-1].startswith(("[from] move: ", "[from]move: ")):
                overridden_move = event.pop().split(": ")[-1]

                if overridden_move == "Sleep Talk":
                    pass
                elif overridden_move in {
                    "Copycat",
                    "Metronome",
                    "Nature Power",
                    "Round",
                }:
                    # triggers moves not owned by actor, so no reveal
                    reveal = False
                elif overridden_move in {"Grass Pledge", "Water Pledge", "Fire Pledge"}:
                    overridden_move = None
                elif self.logger is not None:
                    self.logger.warning(
                        "Unmanaged [from] move message received - move %s in cleaned up "
                        "message %s in battle %s turn %d",
                        overridden_move,
                        event,
                        self.battle_tag,
                        self.turn,
                    )

            if event[-1] == "null":
                event = event[:-1]

            if event[-1].startswith(("[from] ability: ", "[from]ability: ")):
                revealed_ability = event.pop().split(": ")[-1]

                pokemon = event[2]
                self.get_pokemon(pokemon).ability = revealed_ability

                if revealed_ability == "Magic Bounce":
                    use = False
                    reveal = False
                elif revealed_ability == "Dancer":
                    return
                elif self.logger is not None:
                    self.logger.warning(
                        "Unmanaged [from] ability: message received - ability %s in "
                        "cleaned up message %s in battle %s turn %d",
                        revealed_ability,
                        event,
                        self.battle_tag,
                        self.turn,
                    )
            if event[-1] == "[from] Magic Coat" or event[-1] == "[from] Mirror Move":
                use = False
                reveal = False
                event = event[:-1]

            while event[-1] == "[still]":
                event = event[:-1]

            presumed_target = None
            if len(event) == 4:
                pokemon, move = event[2:4]
            elif len(event) == 5:
                pokemon, move, presumed_target = event[2:5]
                if presumed_target == "":
                    pass
                elif len(presumed_target) > 4 and presumed_target[:4] in {
                    "p1: ",
                    "p2: ",
                    "p1a:",
                    "p1b:",
                    "p2a:",
                    "p2b:",
                }:
                    pass
                elif self.logger is not None:
                    self.logger.warning(
                        "Unmanaged move message format received - cleaned up message %s"
                        " in battle %s turn %d",
                        event,
                        self.battle_tag,
                        self.turn,
                    )
            else:
                pokemon, move, presumed_target = event[2:5]
                if (
                    presumed_target == ""
                ):  # ['', 'move', 'p2a: 07ffb4c367', 'Teeter Dance', '', '[from] ability: Dancer']
                    pass
                elif self.logger is not None:
                    self.logger.warning(
                        "Unmanaged move message format received - cleaned up message %s in "
                        "battle %s turn %d",
                        event,
                        self.battle_tag,
                        self.turn,
                    )

            # Check if a silent-effect move has occurred (Minimize) and add the effect
            if move.upper().strip() == "MINIMIZE":
                temp_pokemon = self.get_pokemon(pokemon)
                temp_pokemon.start_effect("MINIMIZE")

            if spread or presumed_target == "":
                presumed_target = None
            pressure = self._pressure_on(pokemon, move, presumed_target)
            mon = self.get_pokemon(pokemon)
            if overridden_move:
                mon.moved(move, failed=failed, use=False, reveal=reveal)
                overridden = mon.moves[Move.retrieve_id(overridden_move)]
                overridden.use(pressure, overridden=True)
            elif not failed and move in {
                "Sleep Talk",
                "Copycat",
                "Metronome",
                "Nature Power",
            }:
                # make preemptive deduction in case override move fails
                mon.moved(move, failed=failed, use=use, reveal=reveal)
            else:
                mon.moved(
                    move, failed=failed, use=use, reveal=reveal, pressure=pressure
                )
        elif event[1] == "cant":
            pokemon, _ = event[2:4]
            self.get_pokemon(pokemon).cant_move()
        elif event[1] == "turn":
            self.end_turn(int(event[2]))
        elif event[1] == "-heal":
            pokemon, hp_status = event[2:4]
            self.get_pokemon(pokemon).set_hp_status(hp_status)
            self._check_heal_message_for_ability(event)
            self._check_heal_message_for_item(event)
        elif event[1] == "-boost":
            pokemon, stat, amount = event[2:5]
            self.get_pokemon(pokemon).boost(stat, int(amount))
        elif event[1] == "-weather":
            weather = event[2]
            if weather == "none":
                self._weather = {}
                return
            else:
                self._weather = {Weather.from_showdown_message(weather): self.turn}
        elif event[1] == "faint":
            mon = self.get_pokemon(event[2])
            mon.faint()
            if mon.species == "dondozo" and isinstance(self.active_pokemon, list):
                active_mons = (
                    self.active_pokemon
                    if event[2][:2] == self.player_role
                    else self.opponent_active_pokemon
                )
                other = active_mons[1 if event[2][2] == "a" else 0]
                if other is not None and Effect.COMMANDER in other.effects:
                    other.end_effect("Commander")
        elif event[1] == "-unboost":
            pokemon, stat, amount = event[2:5]
            self.get_pokemon(pokemon).boost(stat, -int(amount))
        elif event[1] == "-ability":
            pokemon, cause = event[2:4]
            mon = self.get_pokemon(pokemon)
            # As One is a special case ability that combines two abilities
            if "calyrex" in mon.base_species and cause in [
                "As One",
                "Unnerve",
                "Chilling Neigh",
                "Grim Neigh",
            ]:
                return
            if (len(event) > 4 and event[4].startswith("[from] ability: Trace")) or (
                len(event) > 5 and event[5].startswith("[from] ability: Trace")
            ):
                if mon.ability != "trace":
                    # correcting for bad PS ordering of logs, eg:
                    # |-ability|p1a: Gardevoir|Intimidate|boost
                    # |-ability|p1a: Gardevoir|Intimidate|[from] ability: Trace|[of] p2a: Luxray
                    if mon.temporary_ability is not None:
                        mon.temporary_ability = None
                    elif mon.ability is not None:
                        mon._ability = None
                    mon.ability = "trace"
                mon.ability = cause
            elif cause == "Neutralizing Gas":
                self.field_start(cause)
            else:
                mon.ability = cause
        elif split_message[1] == "-start":
            pokemon, effect = event[2:4]
            mon = self.get_pokemon(pokemon)

            if effect == "typechange":
                if len(event) > 5 and event[5].startswith("[of] "):
                    types = "/".join(
                        map(lambda x: x.name, self.get_pokemon(event[5][5:]).types)
                    )
                else:
                    types = event[4]
                mon.start_effect(effect, details=types)
            else:
                if effect == "Mimic":
                    mon._moves.mimic_move = Move(
                        Move.retrieve_id(event[4]), gen=self.gen, from_mimic=True
                    )
                mon.start_effect(effect)

            if mon.is_dynamaxed:
                if mon in self.team.values() and self._dynamax_turn is None:
                    self._dynamax_turn = self.turn
                    self._used_dynamax = True
                elif (
                    mon in self.opponent_team.values()
                    and self._opponent_dynamax_turn is None
                ):
                    self._opponent_dynamax_turn = self.turn
                    self._opponent_used_dynamax = True
        elif event[1] == "-activate":
            target, effect = event[2:4]
            if target and effect.replace("move: ", "") == "Skill Swap":
                if len(event) > 4:
                    if event[-1].startswith("[of] "):
                        target_mon = self.get_pokemon(target)
                        actor_mon = self.get_pokemon(event[-1].replace("[of] ", ""))
                        abilities = [
                            a.replace("[ability] ", "").replace("[ability2] ", "")
                            for a in event[4:-1]
                            if a.replace("[ability] ", "").replace("[ability2] ", "")
                        ]
                        if len(abilities) >= 2:
                            # Opponent swap, gen 5+: abilities revealed
                            target_mon.start_effect(effect, abilities[:2])
                            actor_mon.temporary_ability = abilities[1]
                        elif (
                            target_mon.ability is not None
                            and actor_mon.ability is not None
                        ):
                            # Ally swap or gen <= 4: swap known abilities
                            target_ability = target_mon.ability
                            target_mon.temporary_ability = actor_mon.ability
                            actor_mon.temporary_ability = target_ability
                    else:
                        # Legacy format: actor is event[4], abilities follow
                        actor_mon = self.get_pokemon(event[4])
                        abilities = [
                            a.replace("[ability] ", "").replace("[ability2] ", "")
                            for a in event[5:]
                            if a.replace("[ability] ", "").replace("[ability2] ", "")
                        ]
                        if len(abilities) >= 2:
                            self.get_pokemon(target).start_effect(effect, abilities[:2])
                            actor_mon.temporary_ability = abilities[1]
            elif effect == "ability: Dancer":
                self.get_pokemon(target)._dancing = True
            elif effect == "ability: Mummy":
                target = (
                    event[5].replace("[of] ", "") if "[of] " in event[5] else event[4]
                )
                self.get_pokemon(target).temporary_ability = "mummy"
            elif effect == "ability: Wandering Spirit":
                actor = event[2]
                target = event[6].replace("[of] ", "")
                self.get_pokemon(actor).temporary_ability = event[4]
                self.get_pokemon(target).temporary_ability = "wanderingspirit"
            elif effect == "ability: Symbiosis":
                self.get_pokemon(event[5].replace("[of] ", "")).item = event[4].replace(
                    "[item] ", ""
                )
                self.get_pokemon(target).item = None
            elif effect == "item: Leppa Berry":
                mon = self.get_pokemon(target)
                mv = mon.moves[to_id_str(event[4])]
                # Don't let current pp exceed max pp
                mv._current_pp = min(mv._current_pp + 10, mv.max_pp)
            elif effect == "move: Mimic":
                mon = self.get_pokemon(target)
                mon._moves.mimic_move = Move(
                    Move.retrieve_id(event[4]), gen=self.gen, from_mimic=True
                )
            elif effect == "move: Trick":
                mon = self.get_pokemon(target)
                mon2 = self.get_pokemon(event[4].replace("[of] ", ""))
                mon._item, mon2._item = mon2.item, mon.item
            elif target != "":  # ['', '-activate', '', 'move: Splash']
                self.get_pokemon(target).start_effect(effect)
        elif event[1] == "-status":
            pokemon, status = event[2:4]
            self.get_pokemon(pokemon).status = status
        elif event[1] == "rule":
            self.rules.append(event[2])

        elif event[1] == "-clearallboost":
            self.clear_all_boosts()
        elif event[1] == "-clearboost":
            pokemon = event[2]
            self.get_pokemon(pokemon).clear_boosts()
        elif event[1] == "-clearnegativeboost":
            pokemon = event[2]
            self.get_pokemon(pokemon).clear_negative_boosts()
        elif event[1] == "-clearpositiveboost":
            pokemon = event[2]
            self.get_pokemon(pokemon).clear_positive_boosts()
        elif event[1] == "-copyboost":
            source, target = event[2:4]
            self.get_pokemon(target).copy_boosts(self.get_pokemon(source))
        elif event[1] == "-curestatus":
            pokemon, status = event[2:4]
            self.get_pokemon(pokemon).cure_status(status)
        elif event[1] == "-cureteam":
            pokemon = event[2]
            team = (
                self.team if pokemon[:2] == self._player_role else self._opponent_team
            )
            for mon in team.values():
                mon.cure_status()
        elif event[1] == "-end":
            pokemon, effect = event[2:4]
            if effect == "ability: Neutralizing Gas":
                self._field_end(effect)
            else:
                self.get_pokemon(pokemon).end_effect(effect)
        elif event[1] == "-endability":
            pokemon = event[2]
            self.get_pokemon(pokemon).temporary_ability = None
        elif event[1] == "-enditem":
            pokemon, item = event[2:4]
            self.get_pokemon(pokemon).end_item(item)
        elif event[1] == "-fieldend":
            condition = event[2]
            self._field_end(condition)
        elif event[1] == "-fieldstart":
            condition = event[2]
            self.field_start(condition)
        elif event[1] in ["-formechange", "detailschange"]:
            pokemon, species = event[2:4]
            self.get_pokemon(pokemon).forme_change(species)
        elif event[1] == "-invertboost":
            pokemon = event[2]
            self.get_pokemon(pokemon).invert_boosts()
        elif event[1] == "-item":
            if len(event) == 6:
                item, cause, pokemon = event[3:6]

                if cause == "[from] ability: Frisk":
                    pokemon = pokemon.split("[of] ")[-1]
                    mon = self.get_pokemon(pokemon)

                    if isinstance(self.active_pokemon, list):
                        self.get_pokemon(event[2]).item = to_id_str(item)
                    else:
                        if mon == self.active_pokemon:
                            self.opponent_active_pokemon.item = to_id_str(item)
                        elif mon == self.opponent_active_pokemon:
                            self.active_pokemon.item = to_id_str(item)

                    mon.ability = "frisk"
                elif cause == "[from] ability: Pickpocket":
                    pickpocket = event[2]
                    pickpocketed = event[5].replace("[of] ", "")
                    item = event[3]

                    self.get_pokemon(pickpocket).item = to_id_str(item)
                    self.get_pokemon(pickpocket).ability = "pickpocket"
                    self.get_pokemon(pickpocketed).item = None
                elif cause == "[from] ability: Magician":
                    magician = event[2]
                    victim = event[5].replace("[of] ", "")
                    item = event[3]

                    self.get_pokemon(magician).item = to_id_str(item)
                    self.get_pokemon(magician).ability = "magician"
                    self.get_pokemon(victim).item = None
                elif cause in {"[from] move: Thief", "[from] move: Covet"}:
                    thief = event[2]
                    victim = event[5].replace("[of] ", "")
                    item = event[3]

                    self.get_pokemon(thief).item = to_id_str(item)
                    self.get_pokemon(victim).item = None
                else:
                    raise ValueError(f"Unhandled item message: {event}")
            else:
                pokemon, item = event[2:4]
                if len(event) > 4 and event[4] in [
                    "[from] ability: Magician",
                    "[from] move: Switcheroo",
                    "[from] move: Trick",
                ]:
                    # this event is handled when consuming -activate event
                    return
                self.get_pokemon(pokemon).item = to_id_str(item)
        elif event[1] == "-mega":
            assert self.player_role is not None
            if event[2].startswith(self.player_role):
                self._used_mega_evolve = True
            else:
                self._opponent_used_mega_evolve = True
            pokemon, megastone = event[2:4]
            self.get_pokemon(pokemon).mega_evolve(megastone)
        elif event[1] == "-mustrecharge":
            pokemon = event[2]
            self.get_pokemon(pokemon).must_recharge = True
        elif event[1] == "-prepare":
            try:
                attacker, move, defender = event[2:5]
                defender_mon = (
                    self.get_pokemon(defender) if defender != "[premajor]" else None
                )
                if defender_mon is not None and to_id_str(move) == "skydrop":
                    defender_mon.start_effect("Sky Drop")
            except ValueError:
                attacker, move = event[2:4]
                defender_mon = None
            self.get_pokemon(attacker).prepare(move, defender_mon)
        elif event[1] == "-primal":
            pokemon = event[2]
            self.get_pokemon(pokemon).primal()
        elif event[1] == "-setboost":
            pokemon, stat, amount = event[2:5]
            self.get_pokemon(pokemon).set_boost(stat, int(amount))
        elif event[1] == "-sethp":
            pokemon, hp_status = event[2:4]
            self.get_pokemon(pokemon).set_hp(hp_status)
        elif event[1] == "-sideend":
            side, condition = event[2:4]
            self.side_end(side, condition)
        elif event[1] == "-sidestart":
            side, condition = event[2:4]
            self._side_start(side, condition)
        elif event[1] in ["-singleturn", "-singlemove"]:
            pokemon, effect = event[2:4]
            self.get_pokemon(pokemon).start_effect(effect.replace("move: ", ""))
        elif event[1] == "-swapboost":
            source, target, stats = event[2:5]
            source_mon = self.get_pokemon(source)
            target_mon = self.get_pokemon(target)
            if "[from]" in stats:
                all_stats = ["accuracy", "atk", "def", "evasion", "spa", "spd", "spe"]
                for stat in all_stats:
                    source_mon.boosts[stat], target_mon.boosts[stat] = (
                        target_mon.boosts[stat],
                        source_mon.boosts[stat],
                    )
            else:
                for stat in stats.split(", "):
                    source_mon.boosts[stat], target_mon.boosts[stat] = (
                        target_mon.boosts[stat],
                        source_mon.boosts[stat],
                    )
        elif event[1] == "-transform":
            pokemon, into = event[2:4]
            mon = self.get_pokemon(pokemon)
            if len(event) > 4 and event[4] == "[from] ability: Imposter":
                mon._add_move("transform")
                mon.ability = "imposter"
            mon.transform(self.get_pokemon(into))
        elif event[1] == "-zpower":
            assert self.player_role is not None
            if event[2].startswith(self.player_role):
                self._used_z_move = True
            else:
                self._opponent_used_z_move = True
            pokemon = event[2]
        elif event[1] == "clearpoke":
            self.in_team_preview = True
            for mon in self.team.values():
                mon.clear_active()
        elif event[1] == "gen":
            if self._gen != int(event[2]):
                err = f"Battle Initiated with gen {self._gen} but got: {event}"
                raise RuntimeError(err)
        elif event[1] == "tier":
            self._format = re.sub("[^a-z0-9]+", "", event[2].lower())
        elif event[1] == "inactive":
            if "disconnected" in event[2]:
                self._anybody_inactive = True
            elif "reconnected" in event[2]:
                self._anybody_inactive = False
                self._reconnected = True
        elif event[1] == "player":
            if len(event) == 6:
                player, username, avatar, rating = event[2:6]
            elif len(event) == 5:
                player, username, avatar = event[2:5]
                rating = None
            elif len(event) == 4:
                if event[-1] != "":
                    raise RuntimeError(f"Invalid player message: {event}")
                return
            else:
                if not self._anybody_inactive:
                    if self._reconnected:
                        self._reconnected = False
                    else:
                        raise RuntimeError(f"Invalid player message: {event}")
                return
            if username == self._player_username:
                self._player_role = player
            else:
                self._player_role = "p1" if player == "p2" else "p2"
            if rating is not None:
                return self._players.append(
                    {
                        "username": username,
                        "player": player,
                        "avatar": avatar,
                        "rating": rating,
                    }
                )
            else:
                return self._players.append(
                    {"username": username, "player": player, "avatar": avatar}
                )

        elif event[1] == "poke":
            player, details = event[2:4]
            self._register_teampreview_pokemon(player, details)
        elif event[1] == "raw":
            rating_splint_event = event[2].split("'s rating: ")

            if len(rating_splint_event) != 2:
                return

            username, rating_info = event[2].split("'s rating: ")
            rating_int = int(rating_info[:4])
            if username == self.player_username:
                self._rating = rating_int
            elif username == self.opponent_username:
                self._opponent_rating = rating_int
            elif self.logger is not None:
                self.logger.warning(
                    "Rating information regarding an unrecognized username received. "
                    "Received '%s', while only known players are '%s' and '%s'",
                    username,
                    self.player_username,
                    self.opponent_username,
                )
        elif event[1] == "replace":
            pokemon = event[2]
            details = event[3]
            self.end_illusion(pokemon, details)
        elif event[1] == "start":
            self.in_team_preview = False
        elif event[1] == "swap":
            pokemon, position = event[2:4]
            self._swap(pokemon, position)
        elif event[1] == "teamsize":
            player, number = event[2:4]
            self._team_size[player] = int(number)
        elif event[1] in {"message", "-message"}:
            if self.logger is not None:
                self.logger.info("Received message: %s", event[2])
        elif event[1] == "-immune":
            if len(event) == 4:
                pokemon, cause = event[2:]

                if cause.startswith("[from] ability:"):
                    cause = cause.replace("[from] ability:", "")
                    self.get_pokemon(pokemon).ability = cause
        elif event[1] == "-swapsideconditions":
            self._side_conditions, self._opponent_side_conditions = (
                self._opponent_side_conditions,
                self._side_conditions,
            )
        elif event[1] == "title":
            player_1, player_2 = event[2].split(" vs. ")
            self.players = player_1, player_2
        elif event[1] == "-terastallize":
            pokemon, type_ = event[2:]
            mon = self.get_pokemon(pokemon)
            mon.terastallize(type_)

            if mon.is_terastallized:
                if mon in self.team.values():
                    self._used_tera = True
                elif mon in self.opponent_team.values():
                    self._opponent_used_tera = True
        else:
            raise NotImplementedError(event)

    @abstractmethod
    def parse_request(
        self, request: Dict[str, Any], strict_battle_tracking: bool = False
    ):
        pass

    def _pressure_on(self, pokemon: str, move: str, target_str: Optional[str]) -> bool:
        move_id = Move.retrieve_id(move)
        if move_id not in GenData.from_gen(self.gen).moves:
            # This happens when `move` is a z-move. Since z-moves cannot be PP tracked
            # anyway, we just return False here.
            return False
        move_data = GenData.from_gen(self.gen).moves[move_id]
        target = self._get_target_mon(pokemon, move_data["target"], target_str)
        if target is None:
            return False
        return (
            target.ability == "pressure"
            and not target.fainted
            and (
                move_data["target"]
                in [
                    "all",
                    "allAdjacent",
                    "allAdjacentFoes",
                    "any",
                    "normal",
                    "randomNormal",
                    "scripted",
                ]
                or "mustpressure" in move_data["flags"]
            )
        )

    def _register_teampreview_pokemon(self, player: str, details: str):
        if player != self._player_role:
            mon = Pokemon(details=details, gen=self.gen)
            self._teampreview_opponent_team.append(mon)

    def side_end(self, side: str, condition_str: str):
        if side[:2] == self._player_role:
            conditions = self.side_conditions
        else:
            conditions = self.opponent_side_conditions
        condition = SideCondition.from_showdown_message(condition_str)
        if condition is not SideCondition.UNKNOWN:
            conditions.pop(condition)

    def _side_start(self, side: str, condition_str: str):
        if side[:2] == self._player_role:
            conditions = self.side_conditions
        else:
            conditions = self.opponent_side_conditions
        condition = SideCondition.from_showdown_message(condition_str)
        if condition in STACKABLE_CONDITIONS:
            conditions[condition] = conditions.get(condition, 0) + 1
        elif condition not in conditions:
            conditions[condition] = self.turn

    def _swap(self, pokemon_str: str, slot: str):
        if self.logger is not None:
            self.logger.warning("swap method in Battle is not implemented")

    @abstractmethod
    def switch(self, pokemon_str: str, details: str, hp_status: str):
        pass

    def tied(self):
        if not self._has_terminal_replay_result():
            self._replay_data.append(["", "tie"])
        self._finish_battle()

    def apply_teambuilder_team(
        self,
        role: str,
        teambuilder_team: List[TeambuilderPokemon],
        teampreview_team: List[Pokemon],
    ):
        for preview_mon in teampreview_team:
            teambuilder_mon = [
                m
                for m in teambuilder_team
                if m.nickname is not None and preview_mon.identifies_as(m.nickname)
            ][0]
            mon = self.get_pokemon(
                f"{role}: {teambuilder_mon.nickname}", details=preview_mon._last_details
            )
            mon._update_from_teambuilder(teambuilder_mon)

    def _update_team_from_request(
        self, side: Dict[str, Any], strict_battle_tracking: bool = False
    ):
        # Detect and fix illusion-induced active/inactive mismatches.
        # The request always contains the truth about which mons are active,
        # but switch events during illusion will have set the wrong mon active.\
        falsely_active = []
        truly_active = []
        for pokemon in side["pokemon"]:
            if pokemon["ident"] not in self.team:
                self.get_pokemon(
                    pokemon["ident"],
                    force_self_team=True,
                    details=pokemon["details"],
                    request=pokemon,
                )
            mon = self.team[pokemon["ident"]]
            if pokemon["active"] and not mon.active:
                truly_active.append(mon)
            elif not pokemon["active"] and mon.active:
                falsely_active.append(mon)

        for illusioned in falsely_active:
            illusioned.was_illusioned(self.fields)
        for illusionist in truly_active:
            illusionist.switch_in()

        for pokemon in side["pokemon"]:
            if pokemon["ident"] in self._team:
                if strict_battle_tracking and "illusion" not in [
                    p.ability for p in self.team.values()
                ]:
                    assert self.player_role is not None
                    self._team[pokemon["ident"]].check_consistency(
                        pokemon, self.player_role
                    )
                self._team[pokemon["ident"]].update_from_request(pokemon)
            else:
                self.get_pokemon(
                    pokemon["ident"],
                    force_self_team=True,
                    details=pokemon["details"],
                    request=pokemon,
                )

    def won_by(self, player_name: str):
        if not self._has_terminal_replay_result():
            self._replay_data.append(["", "win", player_name])
        if player_name == self._player_username:
            self._won = True
        else:
            self._won = False
        self._finish_battle()

    def end_turn(self, turn: int):
        self.turn = turn

        for mon in self.all_active_pokemons:
            if mon:
                mon.end_turn()

    @property
    @abstractmethod
    def active_pokemon(self) -> Any:
        pass

    @property
    @abstractmethod
    def all_active_pokemons(self) -> List[Optional[Pokemon]]:
        pass

    @property
    @abstractmethod
    def available_moves(self) -> Any:
        pass

    @property
    @abstractmethod
    def available_switches(self) -> Any:
        pass

    @property
    def battle_tag(self) -> str:
        """
        :return: The battle identifier.
        :rtype: str
        """
        return self._battle_tag

    @property
    @abstractmethod
    def can_dynamax(self) -> Any:
        pass

    @property
    @abstractmethod
    def can_mega_evolve(self) -> Any:
        pass

    @property
    @abstractmethod
    def can_tera(self) -> Any:
        pass

    @property
    @abstractmethod
    def can_z_move(self) -> Any:
        pass

    @property
    def commanding(self) -> bool:
        """
        :return: Whether commander is active with Dondozo and Tatsugiri on the field
        :rtype: bool
        """
        return self._commanding

    @property
    def dynamax_turns_left(self) -> Optional[int]:
        """
        :return: How many turns of dynamax are left. None if dynamax is not active
        :rtype: int, optional
        """
        if self._dynamax_turn is not None and any(
            map(lambda pokemon: pokemon.is_dynamaxed, self._team.values())
        ):
            return max(3 - (self.turn - self._dynamax_turn), 0)
        return None

    @property
    def fields(self) -> Dict[Field, int]:
        """
        :return: A Dict mapping fields to the turn they have been activated.
        :rtype: Dict[Field, int]
        """
        return self._fields

    @property
    def finished(self) -> bool:
        """
        :return: A boolean indicating whether the battle is finished.
        :rtype: Optional[bool]
        """
        return self._finished

    @property
    @abstractmethod
    def force_switch(self) -> Any:
        pass

    @property
    def format(self) -> Optional[str]:
        """
        :return: The format of the battle, in accordance with Showdown protocol
        :rtype: Optional[str]
        """
        return self._format

    @property
    def gen(self) -> int:
        """
        :return: The generation of the battle; will be the parameter with which the
            the battle was initiated
        :rtype: int
        """
        return self._gen

    @property
    @abstractmethod
    def grounded(self) -> Any:
        pass

    @property
    def last_request(self) -> Dict[str, Any]:
        """
        The last request received from the server. This allows players to track
            rqid and also maintain parallel battle copies for search/inference

        :return: The last request.
        :rtype: Dict[str, Any]
        """
        return self._last_request

    @property
    def lost(self) -> Optional[bool]:
        """
        :return: If the battle is finished, a boolean indicating whether the battle is
            lost. Otherwise None.
        :rtype: Optional[bool]
        """
        return None if self._won is None else not self._won

    @property
    def max_team_size(self) -> Optional[int]:
        """
        :return: The maximum acceptable size of the team to return in teampreview, if
            applicable.
        :rtype: int, optional
        """
        return self._max_team_size

    @property
    @abstractmethod
    def maybe_trapped(self) -> Any:
        pass

    @property
    @abstractmethod
    def opponent_active_pokemon(self) -> Any:
        pass

    @property
    def opponent_dynamax_turns_left(self) -> Optional[int]:
        """
        :return: How many turns of dynamax are left for the opponent's pokemon.
            None if dynamax is not active
        :rtype: int | None
        """
        if self._opponent_dynamax_turn is not None and any(
            map(lambda pokemon: pokemon.is_dynamaxed, self._opponent_team.values())
        ):
            return max(3 - (self.turn - self._opponent_dynamax_turn), 0)
        return None

    @property
    def opponent_role(self) -> Optional[str]:
        """
        :return: Opponent's role in given battle. p1/p2
        :rtype: str, optional
        """
        if self.player_role == "p1":
            return "p2"
        if self.player_role == "p2":
            return "p1"
        return None

    @property
    def opponent_side_conditions(self) -> Dict[SideCondition, int]:
        """
        :return: The opponent's side conditions. Keys are SideCondition objects, values
            are:

            - the number of layers of the SideCondition if the side condition is
                stackable
            - the turn where the SideCondition was setup otherwise
        :rtype: Dict[SideCondition, int]
        """
        return self._opponent_side_conditions

    @property
    def opponent_team(self) -> Dict[str, Pokemon]:
        """
        During teampreview, keys are not definitive: please rely on values.

        :return: The opponent's team. Keys are identifiers, values are pokemon objects.
        :rtype: Dict[str, Pokemon]
        """
        if self._opponent_team:
            return self._opponent_team
        else:
            return {mon.species: mon for mon in self._teampreview_opponent_team}

    @property
    def opponent_used_dynamax(self) -> bool:
        """
        :return: Whether or not opponent's current active pokemon can dynamax
        :rtype: bool
        """
        return self._opponent_used_dynamax

    @property
    def opponent_used_mega_evolve(self) -> bool:
        """
        :return: Whether or not opponent's current active pokemon can mega-evolve
        :rtype: bool
        """
        return self._opponent_used_mega_evolve

    @property
    def opponent_used_tera(self) -> bool:
        """
        :return: Whether or not opponent's current active pokemon can terastallize
        :rtype: bool
        """
        return self._opponent_used_tera

    @property
    def opponent_used_z_move(self) -> bool:
        """
        :return: Whether or not opponent's current active pokemon can z-move
        :rtype: bool
        """
        return self._opponent_used_z_move

    @property
    def opponent_username(self) -> Optional[str]:
        """
        :return: The opponent's username, or None if unknown.
        :rtype: str, optional.
        """
        return self._opponent_username

    @opponent_username.setter
    def opponent_username(self, value: str):
        self._opponent_username = value

    @property
    def player_role(self) -> Optional[str]:
        """
        :return: Player's role in given battle. p1/p2
        :rtype: str, optional
        """
        return self._player_role

    @player_role.setter
    def player_role(self, value: Optional[str]):
        self._player_role = value

    @property
    def player_username(self) -> str:
        """
        :return: The player's username.
        :rtype: str
        """
        return self._player_username

    @player_username.setter
    def player_username(self, value: str):
        self._player_username = value

    @property
    def players(self) -> Tuple[str, str]:
        """
        :return: The pair of players' usernames.
        :rtype: Tuple[str, str]
        """
        return self._players[0]["username"], self._players[1]["username"]

    @players.setter
    def players(self, players: Tuple[str, str]):
        """Set the players' usernames.

        :param players: Tuple containing the player's username and the
            opponent's username.
        :type players: Tuple[str, str]
        """
        player_1, player_2 = players
        if player_1 != self._player_username:
            self._opponent_username = player_1
        else:
            self._opponent_username = player_2

    @property
    def rating(self) -> Optional[int]:
        """
        Player's rating after the end of the battle, if it was received.

        :return: The player's rating after the end of the battle.
        :rtype: int, optional
        """
        return self._rating

    @property
    def opponent_rating(self) -> Optional[int]:
        """
        Opponent's rating after the end of the battle, if it was received.

        :return: The opponent's rating after the end of the battle.
        :rtype: int, optional
        """
        return self._opponent_rating

    @property
    def side_conditions(self) -> Dict[SideCondition, int]:
        """
        :return: The player's side conditions. Keys are SideCondition objects, values
            are:

            - the number of layers of the side condition if the side condition is
                stackable
            - the turn where the SideCondition was setup otherwise
        :rtype: Dict[SideCondition, int]
        """
        return self._side_conditions

    @property
    def team(self) -> Dict[str, Pokemon]:
        """
        :return: The player's team. Keys are identifiers, values are pokemon objects.
        :rtype: Dict[str, Pokemon]
        """
        return self._team

    @team.setter
    def team(self, value: Dict[str, Pokemon]):
        self._team = value

    @property
    def team_size(self) -> int:
        """
        :return: The number of Pokemon in the player's team.
        :rtype: int
        """
        if self._player_role is not None:
            return self._team_size[self._player_role]
        raise ValueError(
            "Team size cannot be inferred without an assigned player role."
        )

    @property
    def teambuilder_team(self) -> list[TeambuilderPokemon] | None:
        return self._teambuilder_team

    @property
    def teampreview(self) -> bool:
        """
        :return: Wheter the battle is awaiting a teampreview order.
        :rtype: bool
        """
        return self._teampreview

    @property
    def teampreview_team(self) -> List[Pokemon]:
        """
        :return: The player's team during teampreview.
        :rtype: List[Pokemon]
        """
        return self._teampreview_team

    @teampreview_team.setter
    def teampreview_team(self, value: List[Pokemon]):
        self._teampreview_team = value

    @property
    def teampreview_opponent_team(self) -> List[Pokemon]:
        """
        :return: The opponent's team during teampreview.
        :rtype: List[Pokemon]
        """
        return self._teampreview_opponent_team

    @property
    @abstractmethod
    def trapped(self) -> Any:
        pass

    @trapped.setter
    @abstractmethod
    def trapped(self, value: Any):
        pass

    @property
    def turn(self) -> int:
        """
        :return: The current battle turn.
        :rtype: int
        """
        return self._turn

    @turn.setter
    def turn(self, turn: int):
        """Sets the current turn counter to given value.

        :param turn: Current turn value.
        :type turn: int
        """
        self._turn = turn

    @property
    def used_dynamax(self) -> bool:
        """
        :return: Whether or not the current active pokemon can dynamax
        :rtype: bool
        """
        return self._used_dynamax

    @property
    def used_mega_evolve(self) -> bool:
        """
        :return: Whether or not the current active pokemon can mega evolve.
        :rtype: bool
        """
        return self._used_mega_evolve

    @property
    def used_tera(self) -> bool:
        """
        :return: Whether or not the current active pokemon can terastallize
        :rtype: bool
        """
        return self._used_tera

    @property
    def used_z_move(self) -> bool:
        """
        :return: Whether or not the current active pokemon can z-move.
        :rtype: bool
        """
        return self._used_z_move

    @property
    @abstractmethod
    def valid_orders(self) -> Any:
        pass

    @property
    def wait(self) -> bool:
        """
        :return: If True, the battle does not currently need an action from the player.
        :rtype: bool
        """
        return self._wait

    @property
    def weather(self) -> Dict[Weather, int]:
        """
        :return: A Dict mapping the battle's weather (if any) to its starting turn
        :rtype: Dict[Weather, int]
        """
        return self._weather

    @property
    def won(self) -> Optional[bool]:
        """
        :return: If the battle is finished, a boolean indicating whether the battle is
            won. Otherwise None.
        :rtype: Optional[bool]
        """
        return self._won

    @property
    def reviving(self) -> bool:
        return self._reviving



"""This module defines the ConstantTeambuilder class, which is a subclass of
ShowdownTeamBuilder that yields a constant team.
"""

from typing import List

from poke_env.teambuilder.teambuilder import Teambuilder
from poke_env.teambuilder.teambuilder_pokemon import TeambuilderPokemon


class ConstantTeambuilder(Teambuilder):
    def __init__(self, team: str):
        if "|" in team:
            self._mons = self.parse_packed_team(team)
        else:
            self._mons = self.parse_showdown_team(team)

        self.packed_team = self.join_team(self._mons)

    def yield_team(self) -> str:
        return self.packed_team

    @property
    def team(self) -> List[TeambuilderPokemon]:
        return self._mons



from pettingzoo.utils.env import ActionType  # type: ignore[import-untyped]

from poke_env.environment.doubles_env import DoublesEnv
from poke_env.environment.env import PokeEnv
from poke_env.environment.single_agent_wrapper import SingleAgentWrapper
from poke_env.environment.singles_env import SinglesEnv

__all__ = ["ActionType", "PokeEnv", "SingleAgentWrapper", "SinglesEnv", "DoublesEnv"]



from poke_env.battle.abstract_battle import AbstractBattle
from poke_env.battle.battle import Battle
from poke_env.battle.double_battle import DoubleBattle
from poke_env.battle.effect import Effect
from poke_env.battle.field import Field
from poke_env.battle.move import SPECIAL_MOVES, Move, MoveSet
from poke_env.battle.move_category import MoveCategory
from poke_env.battle.pokemon import Pokemon
from poke_env.battle.pokemon_gender import PokemonGender
from poke_env.battle.pokemon_type import PokemonType
from poke_env.battle.side_condition import STACKABLE_CONDITIONS, SideCondition
from poke_env.battle.status import Status
from poke_env.battle.target import Target
from poke_env.battle.weather import Weather
from poke_env.battle.z_crystal import Z_CRYSTAL

__all__ = [
    "AbstractBattle",
    "Battle",
    "DoubleBattle",
    "Effect",
    "Field",
    "Move",
    "MoveSet",
    "MoveCategory",
    "Pokemon",
    "PokemonGender",
    "PokemonType",
    "SPECIAL_MOVES",
    "STACKABLE_CONDITIONS",
    "SideCondition",
    "Status",
    "Target",
    "Weather",
    "Z_CRYSTAL",
]



from typing import Optional, Union

import numpy as np
import numpy.typing as npt
from gymnasium.spaces import MultiDiscrete

from poke_env.battle.double_battle import DoubleBattle
from poke_env.battle.move import SPECIAL_MOVES
from poke_env.battle.pokemon import Pokemon
from poke_env.data import GenData
from poke_env.environment.env import PokeEnv
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    DoubleBattleOrder,
    ForfeitBattleOrder,
    PassBattleOrder,
    SingleBattleOrder,
)
from poke_env.player.player import Player
from poke_env.ps_client import (
    AccountConfiguration,
    LocalhostServerConfiguration,
    ServerConfiguration,
)
from poke_env.teambuilder import Teambuilder


class DoublesEnv(PokeEnv[npt.NDArray[np.int64]]):
    def __init__(
        self,
        account_configuration1: Optional[AccountConfiguration] = None,
        account_configuration2: Optional[AccountConfiguration] = None,
        avatar: Optional[int] = None,
        battle_format: str = "gen8randombattle",
        log_level: Optional[int] = None,
        save_replays: Union[bool, str] = False,
        server_configuration: Optional[
            ServerConfiguration
        ] = LocalhostServerConfiguration,
        accept_open_team_sheet: Optional[bool] = False,
        start_timer_on_battle_start: bool = False,
        start_listening: bool = True,
        open_timeout: Optional[float] = 10.0,
        ping_interval: Optional[float] = 20.0,
        ping_timeout: Optional[float] = 20.0,
        challenge_timeout: Optional[float] = 60.0,
        team: Optional[Union[str, Teambuilder]] = None,
        choose_on_teampreview: bool = True,
        fake: bool = False,
        strict: bool = True,
    ):
        super().__init__(
            account_configuration1=account_configuration1,
            account_configuration2=account_configuration2,
            avatar=avatar,
            battle_format=battle_format,
            log_level=log_level,
            save_replays=save_replays,
            server_configuration=server_configuration,
            accept_open_team_sheet=accept_open_team_sheet,
            start_timer_on_battle_start=start_timer_on_battle_start,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            challenge_timeout=challenge_timeout,
            team=team,
            choose_on_teampreview=choose_on_teampreview,
            fake=fake,
            strict=strict,
        )
        gen = GenData.from_format(battle_format).gen
        action_space_size = DoublesEnv.get_action_space_size(gen)
        self.action_spaces = {
            agent: MultiDiscrete([action_space_size, action_space_size])
            for agent in self.possible_agents
        }

    @staticmethod
    def action_to_order(
        action: npt.NDArray[np.int64],
        battle: DoubleBattle,
        fake: bool = False,
        strict: bool = True,
    ) -> BattleOrder:
        """Convert an action array into a :class:`BattleOrder`.

        The action is a list in doubles, and the individual action mapping is
        as follows, where each 5-long range for a move corresponds to a
        different target (-2, -1, 0, 1, 2).

        element = -2: default
        element = -1: forfeit
        element = 0: pass
        1 <= element <= 6: switch
        7 <= element <= 11: move 1
        12 <= element <= 16: move 2
        17 <= element <= 21: move 3
        22 <= element <= 26: move 4
        27 <= element <= 31: move 1 and mega evolve
        32 <= element <= 36: move 2 and mega evolve
        37 <= element <= 41: move 3 and mega evolve
        42 <= element <= 46: move 4 and mega evolve
        47 <= element <= 51: move 1 and z-move
        52 <= element <= 56: move 2 and z-move
        57 <= element <= 61: move 3 and z-move
        62 <= element <= 66: move 4 and z-move
        67 <= element <= 71: move 1 and dynamax
        72 <= element <= 76: move 2 and dynamax
        77 <= element <= 81: move 3 and dynamax
        82 <= element <= 86: move 4 and dynamax
        87 <= element <= 91: move 1 and terastallize
        92 <= element <= 96: move 2 and terastallize
        97 <= element <= 101: move 3 and terastallize
        102 <= element <= 106: move 4 and terastallize

        :param action: The action to take.
        :type action: ndarray[int64]
        :param battle: The current battle state
        :type battle: AbstractBattle
        :param fake: If ``True``, return a best-effort order even if it would be
            illegal.
        :type fake: bool
        :param strict: If ``True``, raise an error when the action is illegal;
            otherwise return a default order.
        :type strict: bool

        :return: The battle order for the given action in context of the current battle.
        :rtype: BattleOrder

        """
        if action[0] == -2 and action[1] == -2:
            return DefaultBattleOrder()
        elif action[0] == -1 or action[1] == -1:
            return ForfeitBattleOrder()
        try:
            order1 = DoublesEnv._action_to_order_individual(action[0], battle, fake, 0)
        except ValueError as e:
            if strict:
                raise e
            else:
                if battle.logger is not None:
                    battle.logger.warning(str(e) + " Defaulting to random move.")
                order = Player.choose_random_doubles_move(battle)
                order1 = (
                    order.first_order
                    if not isinstance(order, DefaultBattleOrder)
                    else order
                )
        try:
            order2 = DoublesEnv._action_to_order_individual(action[1], battle, fake, 1)
        except ValueError as e:
            if strict:
                raise e
            else:
                if battle.logger is not None:
                    battle.logger.warning(str(e) + " Defaulting to random move.")
                order = Player.choose_random_doubles_move(battle)
                order2 = (
                    order.second_order
                    if not isinstance(order, DefaultBattleOrder)
                    else order
                )
        joined_orders = DoubleBattleOrder.join_orders([order1], [order2])
        if not joined_orders:
            error_msg = (
                f"Invalid action {action} from player {battle.player_username} "
                f"in battle {battle.battle_tag} - converted orders {order1} "
                f"and {order2} are incompatible!"
            )
            if strict:
                raise ValueError(error_msg)
            else:
                if battle.logger is not None:
                    battle.logger.warning(error_msg + " Defaulting to random move.")
                return Player.choose_random_doubles_move(battle)
        else:
            return joined_orders[0]

    @staticmethod
    def _action_to_order_individual(
        action: np.int64, battle: DoubleBattle, fake: bool, pos: int
    ) -> SingleBattleOrder:
        if action == -2:
            return DefaultBattleOrder()
        elif action == 0:
            order: SingleBattleOrder = PassBattleOrder()
        elif action < 7:
            order = Player.create_order(list(battle.team.values())[action - 1])
        else:
            active_mon = battle.active_pokemon[pos]
            if active_mon is None:
                raise ValueError(
                    f"Invalid order from player {battle.player_username} "
                    f"in battle {battle.battle_tag} at position {pos} - action "
                    f"specifies a move, but battle.active_pokemon is None!"
                )
            avail_ids = [m.id for m in battle.available_moves[pos]]
            known_moves = list(active_mon.moves.values())[:4]
            known_ids = [m.id for m in known_moves]
            mvs = (
                battle.available_moves[pos]
                if len(avail_ids) == 1 and avail_ids[0] not in known_ids
                else known_moves
            )
            if (action - 7) % 20 // 5 not in range(len(mvs)):
                raise ValueError(
                    f"Invalid action {action} from player {battle.player_username} "
                    f"in battle {battle.battle_tag} at position {pos} - action "
                    f"specifies a move but the move index {(action - 7) % 20 // 5} "
                    f"is out of bounds for available moves {mvs}!"
                )
            order = Player.create_order(
                mvs[(action - 7) % 20 // 5],
                move_target=(action.item() - 7) % 5 - 2,
                mega=(action - 7) // 20 == 1,
                z_move=(action - 7) // 20 == 2,
                dynamax=(action - 7) // 20 == 3,
                terastallize=(action - 7) // 20 == 4,
            )
        if not fake and str(order) not in [str(o) for o in battle.valid_orders[pos]]:
            raise ValueError(
                f"Invalid action {action} from player {battle.player_username} "
                f"in battle {battle.battle_tag} at position {pos} - order {order} "
                f"not in action space {[str(o) for o in battle.valid_orders[pos]]}!"
            )
        return order

    @staticmethod
    def order_to_action(
        order: BattleOrder,
        battle: DoubleBattle,
        fake: bool = False,
        strict: bool = True,
    ) -> npt.NDArray[np.int64]:
        """Convert a :class:`BattleOrder` into an action array.

        :param order: The order to take.
        :type order: BattleOrder
        :param battle: The current battle state
        :type battle: AbstractBattle
        :param fake: If ``True``, return a best-effort action even if it would be
            illegal.
        :type fake: bool
        :param strict: If ``True``, raise an error when the order is illegal;
            otherwise return default.
        :type strict: bool

        :return: The action for the given battle order in context of the current battle.
        :rtype: ndarray[int64]
        """
        if isinstance(order, DefaultBattleOrder):
            return np.array([-2, -2])
        elif isinstance(order, ForfeitBattleOrder):
            return np.array([-1, -1])
        assert isinstance(order, DoubleBattleOrder)
        joined_orders = DoubleBattleOrder.join_orders(
            [order.first_order], [order.second_order]
        )
        if not fake and not joined_orders:
            error_msg = (
                f"Invalid order {order} from player {battle.player_username} "
                f"in battle {battle.battle_tag} - orders are incompatible!"
            )
            if strict:
                raise ValueError(error_msg)
            else:
                if battle.logger is not None:
                    battle.logger.warning(error_msg + " Defaulting to random move.")
                return DoublesEnv.order_to_action(
                    Player.choose_random_doubles_move(battle), battle, fake, strict
                )
        try:
            action1 = DoublesEnv._order_to_action_individual(
                order.first_order, battle, fake, 0
            )
        except ValueError as e:
            if strict:
                raise e
            else:
                if battle.logger is not None:
                    battle.logger.warning(str(e) + " Defaulting to random move.")
                order = Player.choose_random_doubles_move(battle)
                action1 = DoublesEnv._order_to_action_individual(
                    (
                        order.first_order
                        if not isinstance(order, DefaultBattleOrder)
                        else order
                    ),
                    battle,
                    fake,
                    0,
                )
        try:
            action2 = DoublesEnv._order_to_action_individual(
                order.second_order, battle, fake, 1
            )
        except ValueError as e:
            if strict:
                raise e
            else:
                if battle.logger is not None:
                    battle.logger.warning(str(e) + " Defaulting to random move.")
                order = Player.choose_random_doubles_move(battle)
                action2 = DoublesEnv._order_to_action_individual(
                    (
                        order.second_order
                        if not isinstance(order, DefaultBattleOrder)
                        else order
                    ),
                    battle,
                    fake,
                    1,
                )
        return np.array([action1, action2])

    @staticmethod
    def _order_to_action_individual(
        order: SingleBattleOrder, battle: DoubleBattle, fake: bool, pos: int
    ) -> np.int64:
        if isinstance(order.order, str):
            if isinstance(order, DefaultBattleOrder):
                return np.int64(-2)
            else:
                assert isinstance(order, PassBattleOrder)
                return np.int64(0)
        if not fake and str(order) not in [str(o) for o in battle.valid_orders[pos]]:
            raise ValueError(
                f"Invalid order from player {battle.player_username} in battle "
                f"{battle.battle_tag} at position {pos} - order {order} not in "
                f"action space {[str(o) for o in battle.valid_orders[pos]]}!"
            )
        if isinstance(order.order, Pokemon):
            action = [p.base_species for p in battle.team.values()].index(
                order.order.base_species
            ) + 1
        else:
            active_mon = battle.active_pokemon[pos]
            assert active_mon is not None
            avail_ids = [m.id for m in battle.available_moves[pos]]
            known_moves = list(active_mon.moves.values())[:4]
            known_ids = [m.id for m in known_moves]
            mvs = (
                battle.available_moves[pos]
                if len(avail_ids) == 1 and avail_ids[0] not in known_ids
                else known_moves
            )
            action = [m.id for m in mvs].index(order.order.id)
            target = order.move_target + 2
            if order.mega:
                gimmick = 1
            elif order.z_move:
                gimmick = 2
            elif order.dynamax:
                gimmick = 3
            elif order.terastallize:
                gimmick = 4
            else:
                gimmick = 0
            action = 1 + 6 + 5 * action + target + 20 * gimmick
        return np.int64(action)

    @staticmethod
    def get_action_mask(battle: DoubleBattle) -> list[int]:
        action_mask1 = DoublesEnv.get_action_mask_individual(battle, 0)
        action_mask2 = DoublesEnv.get_action_mask_individual(battle, 1)
        return action_mask1 + action_mask2

    @staticmethod
    def get_action_mask_individual(battle: DoubleBattle, pos: int) -> list[int]:
        switch_space = [
            i + 1
            for i, pokemon in enumerate(battle.team.values())
            if not battle.trapped[pos]
            and pokemon.base_species
            in [p.base_species for p in battle.available_switches[pos]]
        ]
        active_mon = battle.active_pokemon[pos]
        if battle._wait or (any(battle.force_switch) and not battle.force_switch[pos]):
            actions = [0]
        elif all(battle.force_switch) and len(battle.available_switches[0]) == 1:
            actions = switch_space + [0]
        elif battle.teampreview:
            actions = [
                i
                for i, p in enumerate(battle.team.values(), start=1)
                if not p.selected_in_teampreview
            ]
        elif active_mon is None:
            actions = switch_space
        else:
            known_moves = list(active_mon.moves.values())[:4]
            move_spaces = [
                [
                    7 + 5 * i + j + 2
                    for j in battle.get_possible_showdown_targets(move, active_mon)
                ]
                for i, move in enumerate(known_moves)
                if move.id in [m.id for m in battle.available_moves[pos]]
            ]
            move_space = [i for s in move_spaces for i in s]
            if (
                not move_space
                and len(battle.available_moves[pos]) == 1
                and battle.available_moves[pos][0].id not in SPECIAL_MOVES
            ):
                move_space = [
                    7 + j + 2
                    for j in battle.get_possible_showdown_targets(
                        battle.available_moves[pos][0], active_mon
                    )
                ]
            avail_move_ids = [m.id for m in battle.available_moves[pos]]
            available_z_ids = [m.id for m in active_mon.available_z_moves]
            mega_space = [i + 20 for i in move_space if battle.can_mega_evolve[pos]]
            zmove_spaces = [
                [
                    47 + 5 * i + j + 2
                    for j in battle.get_possible_showdown_targets(move, active_mon)
                ]
                for i, move in enumerate(known_moves)
                if move.id in avail_move_ids
                and move.id in available_z_ids
                and battle.can_z_move[pos]
            ]
            zmove_space = [i for s in zmove_spaces for i in s]
            dynamax_spaces = [
                [
                    67 + 5 * i + j + 2
                    for j in battle.get_possible_showdown_targets(
                        move, active_mon, dynamax=True
                    )
                ]
                for i, move in enumerate(known_moves)
                if move.id in avail_move_ids and battle.can_dynamax[pos]
            ]
            dynamax_space = [i for s in dynamax_spaces for i in s]
            tera_space = [i + 80 for i in move_space if battle.can_tera[pos]]
            if (
                not move_space
                and len(battle.available_moves[pos]) == 1
                and battle.available_moves[pos][0].id in SPECIAL_MOVES
            ):
                move_space = [9]
            actions = (
                switch_space
                + move_space
                + mega_space
                + zmove_space
                + dynamax_space
                + tera_space
            )
        actions = actions or [0]
        action_mask = [
            int(i in actions)
            for i in range(DoublesEnv.get_action_space_size(battle.gen))
        ]
        return action_mask

    @staticmethod
    def get_action_space_size(gen: int) -> int:
        num_switches = 6
        num_moves = 4
        num_targets = 5
        if gen == 6:
            num_gimmicks = 1
        elif gen == 7:
            num_gimmicks = 2
        elif gen == 8:
            num_gimmicks = 3
        elif gen == 9:
            num_gimmicks = 4
        else:
            num_gimmicks = 0
        return 1 + num_switches + num_moves * num_targets * (num_gimmicks + 1)



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import List, Union

from poke_env.battle.move import Move
from poke_env.battle.pokemon import Pokemon


class BattleOrder:
    def __str__(self) -> str:
        return self.message

    @property
    @abstractmethod
    def message(self) -> str:
        pass

    def __eq__(self, other) -> bool:
        return isinstance(other, BattleOrder) and self.message == other.message


class ForfeitBattleOrder(BattleOrder):
    @property
    def message(self) -> str:
        return "/forfeit"


class _EmptyBattleOrder(BattleOrder):
    @property
    def message(self) -> str:
        return ""


@dataclass
class SingleBattleOrder(BattleOrder):
    order: Union[Move, Pokemon, str]
    mega: bool = False
    z_move: bool = False
    dynamax: bool = False
    terastallize: bool = False
    move_target: int = 0

    @property
    def message(self) -> str:
        if isinstance(self.order, Move):
            if self.order.id == "recharge":
                return "/choose move 1"

            message = f"/choose move {self.order.id}"
            if self.mega:
                message += " mega"
            elif self.z_move:
                message += " zmove"
            elif self.dynamax:
                message += " dynamax"
            elif self.terastallize:
                message += " terastallize"

            if self.move_target != 0:
                message += f" {self.move_target}"
            return message
        elif isinstance(self.order, Pokemon):
            return f"/choose switch {self.order.name}"
        else:
            return self.order


class PassBattleOrder(SingleBattleOrder):
    def __init__(self):
        super().__init__("/choose pass")


@dataclass
class DoubleBattleOrder(BattleOrder):
    first_order: SingleBattleOrder = field(default_factory=PassBattleOrder)
    second_order: SingleBattleOrder = field(default_factory=PassBattleOrder)

    @property
    def message(self) -> str:
        return f"{self.first_order.message}, {self.second_order.message[8:]}"

    @staticmethod
    def join_orders(
        first_orders: List[SingleBattleOrder], second_orders: List[SingleBattleOrder]
    ) -> List[DoubleBattleOrder]:
        return [
            DoubleBattleOrder(first_order=first_order, second_order=second_order)
            for first_order in first_orders or [PassBattleOrder()]
            for second_order in second_orders or [PassBattleOrder()]
            if not (first_order.mega and second_order.mega)
            if not (first_order.z_move and second_order.z_move)
            if not (first_order.dynamax and second_order.dynamax)
            if not (first_order.terastallize and second_order.terastallize)
            if not (
                isinstance(first_order.order, Pokemon)
                and isinstance(second_order.order, Pokemon)
                and str(first_order) == str(second_order)
            )
            if not (
                isinstance(first_order, PassBattleOrder)
                and isinstance(second_order, PassBattleOrder)
            )
        ]


class DefaultBattleOrder(SingleBattleOrder):
    def __init__(self):
        super().__init__("/choose default")



"""This module defines a gen1-2 damage calculator"""

import math
from typing import List, Optional, Union

from poke_env.battle import (
    Battle,
    DoubleBattle,
    Effect,
    Field,
    Move,
    MoveCategory,
    Pokemon,
    PokemonType,
    SideCondition,
    Status,
    Weather,
)
from poke_env.data import GenData

# Source: https://github.com/smogon/damage-calc/blob/master/calc/src/mechanics/gen12.ts


def calculate_damage_gen12(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    is_critical: bool = False,
):
    """Return the possible damage range for a move.

    :param attacker_identifier: Identifier of the attacking Pokémon.
    :type attacker_identifier: str
    :param defender_identifier: Identifier of the defending Pokémon.
    :type defender_identifier: str
    :param move: Move being used.
    :type move: Move
    :param battle: Current battle object containing both Pokémon.
    :type battle: Battle or DoubleBattle
    :param is_critical: Whether to compute damage for a critical hit.
    :type is_critical: bool
    :return: Tuple of minimum, maximum, and complete list of damage rolls.
    :rtype: Tuple[int, int, list]
    """
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)
    assert battle.player_role is not None and battle.opponent_role is not None
    assert all(
        map(
            lambda x: isinstance(x, int) or isinstance(x, float),
            attacker.stats.values(),
        )
    ), "attacker stats not defined"
    assert all(
        map(
            lambda x: isinstance(x, int) or isinstance(x, float),
            defender.stats.values(),
        )
    ), "defender stats not defined"
    move_category = move.category
    if move_category == MoveCategory.STATUS:
        return (0, 0, [0])

    if any(map(lambda x: x in PROTECT_EFFECTS, defender.effects)):
        return (0, 0, [0])

    if move.id == "painsplit":
        attacker_hp_stat = attacker.stats.get("hp", None)
        defender_hp_stat = defender.stats.get("hp", None)
        if attacker_hp_stat is not None and defender_hp_stat is not None:
            attacker_hp = (
                math.floor(attacker_hp_stat * attacker.current_hp_fraction)
                if attacker_hp_stat
                else 0
            )
            defender_hp = (
                math.floor(defender_hp_stat * defender.current_hp_fraction)
                if defender_hp_stat
                else 0
            )
            damage = max(0, defender_hp - math.floor((attacker_hp + defender_hp) / 2))
            return (damage, damage, [damage])
        else:
            return (0, 0, [0])
    if battle.gen == 1 and "damage" in move.entry:
        damage = handleFixedDamageMoves(attacker, move)
        return (damage, damage, [damage])

    first_defender_type = defender.type_1
    second_defender_type = defender.type_2 if len(defender.types) > 1 else None
    if (
        second_defender_type
        and first_defender_type != second_defender_type
        and battle.gen == 2
    ):
        first_type_precedence = TYPE_EFFECTIVENESS_PRECEDENCE_RULES.index(
            first_defender_type.name
        )
        second_type_precedence = TYPE_EFFECTIVENESS_PRECEDENCE_RULES.index(
            second_defender_type.name
        )

        if first_type_precedence > second_type_precedence:
            first_defender_type, second_defender_type = (
                second_defender_type,
                first_defender_type,
            )

    move_type = move.type
    first_type_effectiveness = get_move_effectiveness(
        battle.gen,
        move,
        move_type,
        first_defender_type,
        Effect.FORESIGHT in defender.effects,
        Field.GRAVITY in battle.fields,
        defender.item == "ringtarget",
    )
    second_type_effectiveness = (
        1
        if second_defender_type is None
        else get_move_effectiveness(
            battle.gen,
            move,
            move_type or move.type,
            second_defender_type,  # type: ignore
            Effect.FORESIGHT in defender.effects,
            Field.GRAVITY in battle.fields,
            defender.item == "ringtarget",
        )
    )

    type_effectiveness = first_type_effectiveness * second_type_effectiveness
    if type_effectiveness == 0:
        return (0, 0, [0])

    if battle.gen == 2 and "damage" in move.entry:
        damage = handleFixedDamageMoves(attacker, move)
        return (damage, damage, [damage])

    if move.id in ("flail", "reversal"):
        is_critical = False
        p = (48 * attacker.current_hp) // attacker.max_hp
        if p <= 1:
            move._base_power_override = 200
        elif p <= 4:
            move._base_power_override = 150
        elif p <= 9:
            move._base_power_override = 100
        elif p <= 16:
            move._base_power_override = 80
        elif p <= 32:
            move._base_power_override = 40
        else:
            move._base_power_override = 20
    elif move.id == "present":  # doesn't calculate for other base power options
        move._base_power_override = 40
    if move.base_power == 0:
        return (0, 0, [0])

    is_physical = move.category == MoveCategory.PHYSICAL
    attack_stat = "atk" if is_physical else "spa"
    defense_stat = "def" if is_physical else "spd"

    # Stats are guaranteed to be int/float by assertions above
    attacker_attack = attacker.stats[attack_stat]
    defender_defense = defender.stats[defense_stat]
    assert attacker_attack is not None
    assert defender_defense is not None

    at = int(attacker_attack * BOOST_MULTIPLIERS[attacker.boosts[attack_stat]])
    df = int(defender_defense * BOOST_MULTIPLIERS[defender.boosts[defense_stat]])

    # Determine if modifiers should be ignored due to a critical hit
    ignore_mods = is_critical and (
        battle.gen == 1
        or (
            battle.gen == 2
            and attacker.boosts[attack_stat] <= defender.boosts[defense_stat]
        )
    )

    lv = attacker.level
    assert at is not None, f"{attack_stat} is None in attacker.stats"

    if ignore_mods:
        assert attacker_attack is not None
        assert defender_defense is not None
        at = int(attacker_attack)
        df = int(defender_defense)

        if battle.gen == 1:
            lv *= 2
    else:
        if is_physical and attacker.status == Status.BRN:
            at = at // 2
    assert at is not None, f"{attack_stat} is None in attacker.stats"
    assert df is not None, f"{defense_stat} is None in defender.stats"
    # Explosion and Self-Destruct halve the defender's defense
    if move.id in ["explosion", "selfdestruct"]:
        df = df // 2
    defender_side_conditions = (
        battle.side_conditions
        if defender_identifier.startswith(battle.player_role)
        else battle.opponent_side_conditions
    )
    # Reflect and Light Screen double the defender's defense unless mods are ignored
    if not ignore_mods:
        if is_physical and SideCondition.REFLECT in defender_side_conditions:
            df *= 2
        elif not is_physical and SideCondition.LIGHT_SCREEN in defender_side_conditions:
            df *= 2

    # Item-based attack boosts
    if (
        attacker.species == "pikachu"
        and attacker.item == "lightball"
        and not is_physical
    ) or (
        attacker.species in ["cubone", "marowak"]
        and attacker.item == "thickclub"
        and is_physical
    ):
        at *= 2

    # Gen 1 overflow fix: cap stats to 8-bit range
    if at > 255 or df > 255:
        at = (at // 4) % 256
        df = (df // 4) % 256

    # Gen 2 Present has a glitched damage calculation using the secondary types of the Pokemon
    # for the Attacker's Level and Defender's Defense.
    if move.id == "present":
        lookup = {
            "NORMAL": 0,
            "FIGHTING": 1,
            "FLYING": 2,
            "POISON": 3,
            "GROUND": 4,
            "ROCK": 5,
            "BUG": 7,
            "GHOST": 8,
            "STEEL": 9,
            "???": 19,
            "FIRE": 20,
            "WATER": 21,
            "GRASS": 22,
            "ELECTRIC": 23,
            "PSYCHIC": 24,
            "ICE": 25,
            "DRAGON": 26,
            "DARK": 27,
        }

        at = 10

        attacker_type = (
            attacker.type_2 if attacker.type_2 is not None else attacker.type_1
        )
        defender_type = (
            defender.type_2 if defender.type_2 is not None else defender.type_1
        )

        df = max(lookup.get(attacker_type.name, 1), 1)
        lv = max(lookup.get(defender_type.name, 1), 1)

    if defender.name == "ditto" and defender.item == "metalpowder":
        df = math.floor(df * 1.5)

    base_damage = math.floor(
        math.floor(
            math.floor((2 * lv) / 5 + 2) * max(1, at) * move.base_power / max(1, df)
        )
        / 50
    )
    # Gen 1 handles move.isCrit above by doubling level
    if battle.gen == 2 and is_critical:
        base_damage *= 2

    # Gen 2 item handling
    if attacker.item:
        item_boost_type = get_item_boost_type(attacker.item)
    else:
        item_boost_type = None

    if move.type.name == item_boost_type:
        base_damage = math.floor(base_damage * 1.1)

    base_damage = min(997, base_damage) + 2
    if (Weather.SUNNYDAY in battle.weather and move.type == PokemonType.FIRE) or (
        Weather.RAINDANCE in battle.weather and move.type == PokemonType.WATER
    ):

        base_damage = math.floor(base_damage * 1.5)
    elif (
        (Weather.SUNNYDAY in battle.weather and move.type == PokemonType.WATER)
        or (Weather.RAINDANCE in battle.weather and move.type == PokemonType.FIRE)
        or move.id == "solarbeam"
    ):
        base_damage = math.floor(base_damage / 2)

    if any(move.type == t for t in attacker.types):
        base_damage = math.floor(base_damage * 1.5)

    if battle.gen == 1:
        base_damage = math.floor(base_damage * first_type_effectiveness)
        base_damage = math.floor(base_damage * second_type_effectiveness)
    else:
        base_damage = math.floor(base_damage * type_effectiveness)
    # Flail and Reversal don't use random factor
    if move.id in ("flail", "reversal"):
        return (base_damage, base_damage, [base_damage])

    damage_rolls: List[int] = []

    for i in range(217, 256):
        if battle.gen == 2:
            # In Gen 2, damage is always rounded up to at least 1
            dmg = max(1, math.floor((base_damage * i) / 255))
        else:
            # In Gen 1, skip random factor if base damage is 1
            if base_damage == 1:
                dmg = 1
            else:
                dmg = math.floor((base_damage * i) / 255)
        damage_rolls.append(dmg)
    if "multihit" in move.entry:
        damage_matrix = [damage_rolls]
        min_vector = []
        max_vector = []

        for times in range(1, move.n_hit[-1]):
            hit_damage: List[int] = []
            for damage_multiplier in range(217, 256):
                if battle.gen == 2:
                    new_final_damage = max(
                        1, math.floor((base_damage * damage_multiplier) / 255)
                    )
                else:
                    new_final_damage = (
                        1
                        if base_damage == 1
                        else math.floor((base_damage * damage_multiplier) / 255)
                    )

                hit_damage.append(new_final_damage)
            min_vector.append(min(hit_damage))
            max_vector.append(max(hit_damage))

            damage_matrix.append(hit_damage)
        return (min_vector, max_vector, damage_matrix)

    return (min(damage_rolls), max(damage_rolls), damage_rolls)


BOOST_MULTIPLIERS = {
    -6: 2.0 / 8,
    -5: 2.0 / 7,
    -4: 2.0 / 6,
    -3: 2.0 / 5,
    -2: 2.0 / 4,
    -1: 2.0 / 3,
    0: 1,
    1: 1.5,
    2: 2,
    3: 2.5,
    4: 3,
    5: 3.5,
    6: 4,
}

TYPE_EFFECTIVENESS_PRECEDENCE_RULES = [
    "NORMAL",
    "FIRE",
    "WATER",
    "ELECTRIC",
    "GRASS",
    "ICE",
    "FIGHTING",
    "POISON",
    "GROUND",
    "FLYING",
    "PSYCHIC",
    "BUG",
    "ROCK",
    "GHOST",
    "DRAGON",
    "DARK",
    "STEEL",
]
PROTECT_EFFECTS = {
    Effect.PROTECT,
    Effect.SPIKY_SHIELD,
    Effect.KINGS_SHIELD,
    Effect.BANEFUL_BUNKER,
    Effect.BURNING_BULWARK,
    Effect.OBSTRUCT,
    Effect.MAX_GUARD,
    Effect.SILK_TRAP,
}


def handleFixedDamageMoves(attacker: Pokemon, move: Move):
    if move.entry["damage"] == "level":
        return attacker.level
    else:
        return move.entry["damage"]


def get_move_effectiveness(
    gen: int,
    move: Move,
    move_type: PokemonType,
    type: PokemonType,
    is_ghost_revealed: Optional[bool] = False,
    is_gravity: Optional[bool] = False,
    is_ring_target: Optional[bool] = False,
) -> float:
    if (
        is_ghost_revealed
        and type == PokemonType.GHOST
        and move_type in (PokemonType.NORMAL, PokemonType.FIGHTING)
    ):
        return 1
    elif move.id == "struggle" and gen > 1:
        return 1
    elif is_gravity and type == PokemonType.FLYING and move_type == PokemonType.GROUND:
        return 1
    elif move.id == "freezedry" and type == PokemonType.WATER:
        return 2
    else:
        effectiveness = PokemonType.damage_multiplier(
            move_type, type, type_chart=GenData.from_gen(gen).type_chart
        )
        if effectiveness == 0 and is_ring_target:
            effectiveness = 1
        if move.id == "flyingpress":
            effectiveness *= PokemonType.damage_multiplier(
                PokemonType.FLYING, type, type_chart=GenData.from_gen(gen).type_chart
            )
        return effectiveness


def get_item_boost_type(item: str) -> Optional[str]:
    item_type_map = {
        "dracoplate": "DRAGON",
        "dragonfang": "undefined",
        "dragonscale": "DRAGON",
        "dreadplate": "DARK",
        "blackglasses": "DARK",
        "earthplate": "GROUND",
        "softsand": "GROUND",
        "fistplate": "FIGHTING",
        "blackbelt": "FIGHTING",
        "flameplate": "FIRE",
        "charcoal": "FIRE",
        "icicleplate": "ICE",
        "nevermeltice": "ICE",
        "insectplate": "BUG",
        "silverpowder": "BUG",
        "ironplate": "STEEL",
        "metalcoat": "STEEL",
        "meadowplate": "GRASS",
        "roseincense": "GRASS",
        "miracleseed": "GRASS",
        "mindplate": "PSYCHIC",
        "oddincense": "PSYCHIC",
        "twistedspoon": "PSYCHIC",
        "fairyfeather": "FAIRY",
        "pixieplate": "FAIRY",
        "skyplate": "FLYING",
        "sharpbeak": "FLYING",
        "splashplate": "WATER",
        "seaincense": "WATER",
        "waveincense": "WATER",
        "mysticwater": "WATER",
        "spookyplate": "GHOST",
        "spelltag": "GHOST",
        "stoneplate": "ROCK",
        "rockincense": "ROCK",
        "hardstone": "ROCK",
        "toxicplate": "POISON",
        "poisonbarb": "POISON",
        "zapplate": "ELECTRIC",
        "magnet": "ELECTRIC",
        "silkscarf": "NORMAL",
        "pinkbow": "NORMAL",
        "polkadotbow": "NORMAL",
    }

    return item_type_map.get(item)



"""This module contains objects related to player configuration."""

from __future__ import annotations

import random
import string
from typing import Counter, NamedTuple, Optional

CONFIGURATION_FROM_PLAYER_COUNTER: Counter[str] = Counter()


class AccountConfiguration(NamedTuple):
    """Player configuration object. Represented with a tuple with two entries: username and
    password."""

    username: str
    password: Optional[str]

    @classmethod
    def generate(cls, key: str, rand: bool = False) -> AccountConfiguration:
        """
        Generates an AccountConfiguration instance using a provided key.
        The same key can be used multiple times to create multiple usernames.

        :param key: The base of the username. Will be truncated to be a maximum
            of 18 characters long.
        :param rand: If true, the unique identifier appended to the end will be a
            randomized 5-length alphanumeric string instead of an increment for every
            identical key.
        """
        if rand:
            char_space = string.ascii_lowercase + string.digits
            unique_ident = "".join(random.choices(char_space, k=5))
        else:
            CONFIGURATION_FROM_PLAYER_COUNTER.update([key])
            unique_ident = str(CONFIGURATION_FROM_PLAYER_COUNTER[key])
        username = f"{key} {unique_ident}"
        if len(username) > 18:
            username = f"{key[: 18 - len(username)]} {unique_ident}"
        return cls(username, None)



import random
from typing import List, Tuple

from poke_env.battle.abstract_battle import AbstractBattle
from poke_env.battle.battle import Battle
from poke_env.battle.double_battle import DoubleBattle
from poke_env.battle.effect import Effect
from poke_env.battle.move import Move
from poke_env.battle.move_category import MoveCategory
from poke_env.battle.pokemon import Pokemon
from poke_env.battle.side_condition import SideCondition
from poke_env.battle.target import Target
from poke_env.data import GenData
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    DoubleBattleOrder,
    PassBattleOrder,
    SingleBattleOrder,
)
from poke_env.player.player import Player


class RandomPlayer(Player):
    def choose_move(self, battle: AbstractBattle) -> BattleOrder:
        return self.choose_random_move(battle)


class MaxBasePowerPlayer(Player):
    def choose_move(self, battle: AbstractBattle):
        if self.format_is_doubles:
            return self.choose_doubles_move(battle)  # type: ignore
        else:
            return self.choose_singles_move(battle)

    @staticmethod
    def choose_singles_move(battle: AbstractBattle):
        if battle.available_moves:
            best_move = max(battle.available_moves, key=lambda move: move.base_power)
            return Player.create_order(best_move)
        return Player.choose_random_move(battle)

    @staticmethod
    def choose_doubles_move(battle: DoubleBattle):
        orders: List[SingleBattleOrder] = []
        switched_in = None

        if any(battle.force_switch):
            return Player.choose_random_doubles_move(battle)

        can_target_first_opponent = (
            battle.opponent_active_pokemon[0]
            and not battle.opponent_active_pokemon[0].fainted
        )
        can_target_second_opponent = (
            battle.opponent_active_pokemon[1]
            and not battle.opponent_active_pokemon[1].fainted
        )
        can_double_target = can_target_first_opponent and can_target_second_opponent

        for mon, moves, switches in zip(
            battle.active_pokemon, battle.available_moves, battle.available_switches
        ):
            switches = [s for s in switches if s != switched_in]

            if not mon or mon.fainted:
                orders.append(PassBattleOrder())
                continue
            elif not moves and switches:
                mon_to_switch_in = random.choice(switches)
                orders.append(SingleBattleOrder(mon_to_switch_in))
                switched_in = mon_to_switch_in
                continue
            elif not moves:
                orders.append(DefaultBattleOrder())
                continue

            def move_power_with_double_target(move):
                if move.target in {Target.NORMAL, Target.ANY} or not can_double_target:
                    return move.base_power
                return move.base_power * 1.5

            best_move = max(moves, key=move_power_with_double_target)

            # randomly picks between the two opponents for normal move targeting
            targets = battle.get_possible_showdown_targets(best_move, mon)
            opp_targets = [
                t
                for t in targets
                if t in {battle.OPPONENT_1_POSITION, battle.OPPONENT_2_POSITION}
            ]
            if opp_targets:
                target = random.choice(opp_targets)
            else:
                target = random.choice(targets)

            orders.append(SingleBattleOrder(best_move, move_target=target))

        if orders[0] or orders[1]:
            return DoubleBattleOrder(orders[0], orders[1])

        return Player.choose_random_move(battle)


class PseudoBattle(Battle):
    def __init__(self, battle: DoubleBattle, active_id: int, opp_id: int):
        self._gen = battle.gen
        self._active_pokemon = battle.active_pokemon[active_id]
        self._opponent_active_pokemon = battle.opponent_active_pokemon[opp_id]
        self._team = battle.team
        self._opponent_team = battle.opponent_team
        self._available_moves = battle.available_moves[active_id]
        self._available_switches = battle.available_switches[active_id]
        self._force_switch = battle.force_switch[active_id]
        self._trapped = battle.trapped[active_id]
        self._wait = battle.wait
        self._side_conditions = battle.side_conditions
        self._opponent_side_conditions = battle.opponent_side_conditions
        self._can_mega_evolve = battle.can_mega_evolve[active_id]
        self._can_z_move = battle.can_z_move[active_id]
        self._can_dynamax = battle.can_dynamax[active_id]
        self._can_tera = battle.can_tera[active_id]

    @property
    def active_pokemon(self):
        return self._active_pokemon

    @property
    def opponent_active_pokemon(self):
        return self._opponent_active_pokemon


class SimpleHeuristicsPlayer(Player):
    ENTRY_HAZARDS = {
        "spikes": SideCondition.SPIKES,
        "stealhrock": SideCondition.STEALTH_ROCK,
        "stickyweb": SideCondition.STICKY_WEB,
        "toxicspikes": SideCondition.TOXIC_SPIKES,
    }

    ANTI_HAZARDS_MOVES = {"rapidspin", "defog"}

    SPEED_TIER_COEFICIENT = 0.1
    HP_FRACTION_COEFICIENT = 0.4
    SWITCH_OUT_MATCHUP_THRESHOLD = -2

    @staticmethod
    def _estimate_matchup(mon: Pokemon, opponent: Pokemon):
        score = max([opponent.damage_multiplier(t) for t in mon.types if t is not None])
        score -= max(
            [mon.damage_multiplier(t) for t in opponent.types if t is not None]
        )
        if mon.base_stats["spe"] > opponent.base_stats["spe"]:
            score += SimpleHeuristicsPlayer.SPEED_TIER_COEFICIENT
        elif opponent.base_stats["spe"] > mon.base_stats["spe"]:
            score -= SimpleHeuristicsPlayer.SPEED_TIER_COEFICIENT

        score += mon.current_hp_fraction * SimpleHeuristicsPlayer.HP_FRACTION_COEFICIENT
        score -= (
            opponent.current_hp_fraction * SimpleHeuristicsPlayer.HP_FRACTION_COEFICIENT
        )

        return score

    @staticmethod
    def _should_dynamax(battle: AbstractBattle, n_remaining_mons: int):
        if battle.can_dynamax:
            # Last full HP mon
            if (
                len([m for m in battle.team.values() if m.current_hp_fraction == 1])
                == 1
                and battle.active_pokemon.current_hp_fraction == 1
            ):
                return True
            # Matchup advantage and full hp on full hp
            if (
                SimpleHeuristicsPlayer._estimate_matchup(
                    battle.active_pokemon, battle.opponent_active_pokemon
                )
                > 0
                and battle.active_pokemon.current_hp_fraction == 1
                and battle.opponent_active_pokemon.current_hp_fraction == 1
            ):
                return True
            if n_remaining_mons == 1:
                return True
        return False

    @staticmethod
    def _should_terastallize(battle: Battle, move: Move) -> bool:
        active = battle.active_pokemon
        opp_active = battle.opponent_active_pokemon
        if (
            not battle.can_tera
            or not active
            or not opp_active
            or active.tera_type is None
        ):
            return False
        offensive_tera_score = opp_active.damage_multiplier(move.type)
        defensive_score = min(
            [1 / (active.damage_multiplier(t) or 1 / 8) for t in opp_active.types]
        )
        defensive_tera_score = min(
            [
                1
                / (
                    t.damage_multiplier(
                        active.tera_type,
                        type_chart=GenData.from_gen(battle.gen).type_chart,
                    )
                    or 1 / 8
                )
                for t in opp_active.types
            ]
        )
        return offensive_tera_score * (defensive_tera_score / defensive_score) > 1

    @staticmethod
    def _should_switch_out(battle: AbstractBattle):
        active = battle.active_pokemon
        opponent = battle.opponent_active_pokemon
        # If there is a decent switch in...
        if [
            m
            for m in battle.available_switches
            if SimpleHeuristicsPlayer._estimate_matchup(m, opponent) > 0
        ]:
            # ...and a 'good' reason to switch out
            if active.boosts["def"] <= -3 or active.boosts["spd"] <= -3:
                return True
            if (
                active.boosts["atk"] <= -3
                and active.stats["atk"] >= active.stats["spa"]
            ):
                return True
            if (
                active.boosts["spa"] <= -3
                and active.stats["atk"] <= active.stats["spa"]
            ):
                return True
            if (
                SimpleHeuristicsPlayer._estimate_matchup(active, opponent)
                < SimpleHeuristicsPlayer.SWITCH_OUT_MATCHUP_THRESHOLD
            ):
                return True
        return False

    @staticmethod
    def _stat_estimation(mon: Pokemon, stat: str):
        # Stats boosts value
        if mon.boosts[stat] > 1:
            boost = (2 + mon.boosts[stat]) / 2
        else:
            boost = 2 / (2 - mon.boosts[stat])
        return ((2 * mon.base_stats[stat] + 31) + 5) * boost

    @staticmethod
    def choose_singles_move(battle: Battle) -> Tuple[SingleBattleOrder, float]:
        # Main mons shortcuts
        active = battle.active_pokemon
        opponent = battle.opponent_active_pokemon

        if active is None or opponent is None or battle.wait:
            return Player.choose_random_singles_move(battle), 0

        # Rough estimation of damage ratio
        physical_ratio = SimpleHeuristicsPlayer._stat_estimation(
            active, "atk"
        ) / SimpleHeuristicsPlayer._stat_estimation(opponent, "def")
        special_ratio = SimpleHeuristicsPlayer._stat_estimation(
            active, "spa"
        ) / SimpleHeuristicsPlayer._stat_estimation(opponent, "spd")

        if battle.available_moves and (
            not SimpleHeuristicsPlayer._should_switch_out(battle)
            or not battle.available_switches
        ):
            n_remaining_mons = len(
                [m for m in battle.team.values() if m.fainted is False]
            )
            n_opp_remaining_mons = 6 - len(
                [m for m in battle.opponent_team.values() if m.fainted is True]
            )

            # Entry hazard...
            for move in battle.available_moves:
                # ...setup
                if (
                    n_opp_remaining_mons >= 3
                    and move.id in SimpleHeuristicsPlayer.ENTRY_HAZARDS
                    and SimpleHeuristicsPlayer.ENTRY_HAZARDS[move.id]
                    not in battle.opponent_side_conditions
                ):
                    return Player.create_order(move), 0

                # ...removal
                elif (
                    battle.side_conditions
                    and move.id in SimpleHeuristicsPlayer.ANTI_HAZARDS_MOVES
                    and n_remaining_mons >= 2
                ):
                    return Player.create_order(move), 0

            # Setup moves
            if (
                active.current_hp_fraction == 1
                and SimpleHeuristicsPlayer._estimate_matchup(active, opponent) > 0
            ):
                for move in battle.available_moves:
                    if (
                        move.boosts
                        and sum(move.boosts.values()) >= 2
                        and move.target == "self"
                        and min(
                            [active.boosts[s] for s, v in move.boosts.items() if v > 0]
                        )
                        < 6
                    ):
                        return Player.create_order(move), 0

            move, score = max(
                [
                    (
                        m,
                        m.base_power
                        * (1.5 if m.type in active.types else 1)
                        * (
                            physical_ratio
                            if m.category == MoveCategory.PHYSICAL
                            else special_ratio
                        )
                        * m.accuracy
                        * m.expected_hits
                        * opponent.damage_multiplier(m),
                    )
                    for m in battle.available_moves
                ],
                key=lambda x: x[1],
            )
            return (
                Player.create_order(
                    move,
                    dynamax=SimpleHeuristicsPlayer._should_dynamax(
                        battle, n_remaining_mons
                    ),
                    terastallize=SimpleHeuristicsPlayer._should_terastallize(
                        battle, move
                    ),
                ),
                score,
            )

        if battle.available_switches:
            switches: List[Pokemon] = battle.available_switches
            return (
                Player.create_order(
                    max(
                        switches,
                        key=lambda s: SimpleHeuristicsPlayer._estimate_matchup(
                            s, opponent
                        ),
                    )
                ),
                0,
            )

        return Player.choose_random_singles_move(battle), 0

    @staticmethod
    def get_double_target_multiplier(battle: DoubleBattle, order: SingleBattleOrder):
        can_target_first_opponent = (
            battle.opponent_active_pokemon[0]
            and not battle.opponent_active_pokemon[0].fainted
        )
        can_target_second_opponent = (
            battle.opponent_active_pokemon[1]
            and not battle.opponent_active_pokemon[1].fainted
        )
        can_double_target = can_target_first_opponent and can_target_second_opponent
        return (
            1
            if not hasattr(order, "order")
            or not isinstance(order.order, Move)
            or order.order.target in {Target.NORMAL, Target.ANY}
            or not can_double_target
            else 1.5
        )

    def choose_move(self, battle: AbstractBattle):
        if not isinstance(battle, DoubleBattle):
            return self.choose_singles_move(battle)[0]  # type: ignore
        orders: List[SingleBattleOrder] = []
        for active_id in [0, 1]:
            mon = battle.active_pokemon[active_id]
            if mon is not None and Effect.COMMANDER in mon.effects:
                orders += [PassBattleOrder()]
                continue
            if mon is None and not battle.available_switches[active_id]:
                orders += [PassBattleOrder()]
                continue
            results = [
                self.choose_singles_move(PseudoBattle(battle, active_id, opp_id))
                for opp_id in [0, 1]
            ]
            possible_orders = [r[0] for r in results]
            scores = [r[1] for r in results]
            for order in possible_orders:
                if (
                    order is not None
                    and hasattr(order, "order")
                    and isinstance(order.order, Move)
                    and mon is not None
                ):
                    target = [o for o in possible_orders].index(order) + 1
                    possible_targets = battle.get_possible_showdown_targets(
                        order.order, mon
                    )
                    if target not in possible_targets:
                        target = possible_targets[0]
                    order.move_target = target
            scores = [
                scores[i]
                * self.get_double_target_multiplier(battle, possible_orders[i])
                for i in [0, 1]
            ]
            orders += [
                (
                    max(results, key=lambda a: a[1])[0]
                    if battle.force_switch != [[False, True], [True, False]][active_id]
                    and not (
                        len(battle.available_switches[active_id]) == 1
                        and battle.force_switch == [True, True]
                        and active_id == 1
                    )
                    else PassBattleOrder()
                )
            ]
        joined_orders = DoubleBattleOrder.join_orders([orders[0]], [orders[1]])
        if joined_orders:
            return joined_orders[0]
        else:
            return DoubleBattleOrder(orders[0], DefaultBattleOrder())



"""poke_env.teambuilder module init."""

from poke_env.teambuilder.constant_teambuilder import ConstantTeambuilder
from poke_env.teambuilder.teambuilder import Teambuilder
from poke_env.teambuilder.teambuilder_pokemon import TeambuilderPokemon

__all__ = ["ConstantTeambuilder", "Teambuilder", "TeambuilderPokemon"]



from poke_env.data.gen_data import GenData
from poke_env.data.normalize import to_id_str
from poke_env.data.replay_template import REPLAY_TEMPLATE

__all__ = ["REPLAY_TEMPLATE", "GenData", "to_id_str"]



from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from poke_env.battle.effect import Effect
from poke_env.battle.field import Field
from poke_env.battle.move import SPECIAL_MOVES, Move, MoveSet
from poke_env.battle.pokemon_gender import PokemonGender
from poke_env.battle.pokemon_type import PokemonType
from poke_env.battle.status import Status
from poke_env.battle.target import Target
from poke_env.battle.z_crystal import Z_CRYSTAL
from poke_env.data import GenData, to_id_str
from poke_env.stats import compute_raw_stats
from poke_env.teambuilder.teambuilder_pokemon import TeambuilderPokemon


class Pokemon:
    __slots__ = (
        "_ability",
        "_active",
        "_active_turns",
        "_base_stats",
        "_boosts",
        "_current_hp",
        "_dancing",
        "_effects",
        "_evs",
        "_forme_change_ability",
        "_gen",
        "_gender",
        "_heightm",
        "_item",
        "_ivs",
        "_last_details",
        "_last_request",
        "_level",
        "_max_hp",
        "_moves",
        "_must_recharge",
        "_name",
        "_nature",
        "_possible_abilities",
        "_preparing_move",
        "_preparing_target",
        "_protect_counter",
        "_shiny",
        "_stats",
        "_revealed",
        "_species",
        "_status",
        "_status_counter",
        "_selected_in_teampreview",
        "_temporary_ability",
        "_temporary_base_stats",
        "_temporary_types",
        "_terastallized",
        "_terastallized_type",
        "_type_1",
        "_type_2",
        "_weightkg",
    )

    def __init__(
        self,
        gen: int,
        *,
        species: Optional[str] = None,
        name: Optional[str] = None,
        request_pokemon: Optional[Dict[str, Any]] = None,
        details: Optional[str] = None,
        teambuilder: Optional[TeambuilderPokemon] = None,
    ):
        # Species related attributes
        self._base_stats: Dict[str, int]
        self._heightm: int
        self._possible_abilities: List[str]
        self._species: str = ""
        self._type_1: PokemonType
        self._type_2: Optional[PokemonType] = None
        self._weightkg: int

        # Individual related attributes
        self._ability: Optional[str] = None
        self._gender: Optional[PokemonGender] = None
        self._level: int = 100
        self._max_hp: Optional[int] = 0
        self._moves = MoveSet({})
        self._name: Optional[str] = None
        self._shiny: Optional[bool] = False

        # Teambuilder related attributes
        self._evs: list[int] | None = None
        self._ivs: list[int] | None = None
        self._nature: str | None = None

        # Battle related attributes
        self._gen = gen
        self._active: bool = False
        self._boosts: Dict[str, int] = {
            "accuracy": 0,
            "atk": 0,
            "def": 0,
            "evasion": 0,
            "spa": 0,
            "spd": 0,
            "spe": 0,
        }
        self._current_hp: Optional[int] = 0
        self._effects: Dict[Effect, int] = {}
        self._active_turns: int = 0
        self._terastallized: bool = False
        self._terastallized_type: Optional[PokemonType] = None
        self._item: Optional[str] = GenData.from_gen(gen).UNKNOWN_ITEM
        self._last_request: Optional[Dict[str, Any]] = {}
        self._last_details: str = ""
        self._must_recharge: bool = False
        self._preparing_move: Optional[Move] = None
        self._preparing_target: Optional[bool | Pokemon] = None
        self._protect_counter: int = 0
        self._revealed: bool = False
        self._selected_in_teampreview: bool = False
        self._stats: Dict[str, Optional[int]] = {
            "hp": None,
            "atk": None,
            "def": None,
            "spa": None,
            "spd": None,
            "spe": None,
        }
        self._status: Optional[Status] = None
        self._status_counter: int = 0
        self._temporary_ability: Optional[str] = None
        self._forme_change_ability: Optional[str] = None
        self._temporary_base_stats: Optional[Dict[str, int]] = None
        self._temporary_types: List[PokemonType] = []
        self._dancing = False

        if request_pokemon:
            self.update_from_request(request_pokemon)
        elif details:
            self._update_from_details(details)
        elif species:
            self._update_from_pokedex(species)
        elif teambuilder:
            self._update_from_teambuilder(teambuilder)

        if name is not None:
            self._name = name

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        if self._status is None:
            status_repr = None
        else:
            status_repr = self._status.name

        return (
            f"{self._species} (pokemon object) "
            f"[Active: {self._active}, Status: {status_repr}]"
        )

    def _add_move(self, move_id: str) -> Optional[Move]:
        """Store the move if applicable."""
        id_ = Move.retrieve_id(move_id)
        if id_ in self.moves:
            return self.moves[id_]
        if not Move.should_be_stored(id_, self.gen):
            return None
        if id_ not in self.base_moves:
            move = Move(
                move_id=id_,
                raw_id=move_id,
                gen=self.gen,
                from_transform=self.transformed,
            )
            self.base_moves[id_] = move
        return self.base_moves[id_]

    def boost(self, stat: str, amount: int):
        self._boosts[stat] += amount
        if self._boosts[stat] > 6:
            self._boosts[stat] = 6
        elif self._boosts[stat] < -6:
            self._boosts[stat] = -6

    def cant_move(self):
        self._dancing = False
        self._protect_counter = 0

        if self._status == Status.SLP:
            self._status_counter += 1

    def check_consistency(self, pkmn_request: Dict[str, Any], player_role: str):
        assert (
            pkmn_request["ident"] == f"{player_role}: {self.name}"
        ), f"{pkmn_request['ident']} != {player_role}: {self.name}"
        split_details = pkmn_request["details"].split(", ")
        level = None
        gender = None
        shiny = split_details[-1] == "shiny"
        if shiny:
            split_details.pop()
        if len(split_details) == 3:
            _, level, gender = split_details
        elif len(split_details) == 2:
            if split_details[1].startswith("L"):
                _, level = split_details
            else:
                _, gender = split_details
        level = int(level[1:]) if level is not None else 100
        gender = (
            PokemonGender.from_request_details(gender)
            if gender is not None
            else PokemonGender.NEUTRAL
        )
        assert level == self.level, f"{level} != {self.level}"
        assert self.gender is not None
        assert (
            gender == self.gender
        ), f"{gender.name.lower()} != {self.gender.name.lower()}"
        assert shiny == self.shiny, f"{shiny} != {self.shiny}"
        assert (
            pkmn_request["active"] == self.active
        ), f"{pkmn_request['active']} != {self.active}"
        if self.gen >= 2:
            if self.item == "unknown_item":
                # needed for item initialization in start of game,
                # done anyway in update_from_request()
                self._item = pkmn_request["item"]
            assert pkmn_request["item"] == (
                self.item or ""
            ), f"{pkmn_request['item']} != {self.item or ''}"
        assert len(self.moves) <= 4, f"More than 4 moves: {self.moves}"
        if self.transformed or "transform" in self.moves:
            return
        assert (
            pkmn_request["condition"] == self.hp_status
        ), f"{pkmn_request['condition']} != {self.hp_status}"
        if self.ability is None:
            # needed for ability initialization in start of game,
            # done anyway in update_from_request()
            self.ability = pkmn_request["baseAbility"]
        assert pkmn_request["baseAbility"] == (
            self.base_ability or ""
        ), f"{pkmn_request['baseAbility']} != {self.base_ability or ''}"
        if "ability" in pkmn_request:
            if (
                pkmn_request["baseAbility"] != pkmn_request["ability"]
                and self._temporary_ability is None
            ):
                # needed for ability initialization in start of game,
                # done anyway in update_from_request()
                self._temporary_ability = to_id_str(pkmn_request["ability"])
            assert pkmn_request["ability"] == (
                self.ability or ""
            ), f"{pkmn_request['ability']} != {self.ability or ''}"

    def check_move_consistency(
        self, active_request: Dict[str, Any], is_doubles: bool = False
    ):
        if self.mimic_move is not None and self.mimic_move.id in [
            m.id for m in self.base_moves.values()
        ]:
            return
        for move_request in active_request["moves"]:
            matches = [
                m
                for m in self.moves.values()
                if Move.retrieve_id(m.id) == move_request["id"]
            ]
            if not matches:
                assert (
                    active_request["moves"][0]["id"] in SPECIAL_MOVES
                    or self.ability == "dancer"
                    or {
                        "copycat",
                        "metronome",
                        "mefirst",
                        "mirrormove",
                        "assist",
                    }.intersection(self.moves)
                ), (
                    f"Error with move {move_request['id']} from json request. "
                    "Expected self.moves to contain copycat, metronome, mefirst, "
                    "mirrormove, or assist, or to have the ability dancer. Got "
                    f"{self.moves}, ability: {self.ability}"
                )
                continue
            assert (
                len(matches) == 1
            ), f"Expected exactly one move to match {move_request['id']}. Got {matches}"
            move = matches[0]
            if (
                "pp" in move_request
                and self.gen not in [1, 2, 3, 7, 8]
                and not is_doubles
            ):
                # exclude early gens because of unreliable Showdown event messages
                # exclude gen 7 and 8 because of Z-move and Max Move PP untrackability
                # exclude doubles because targets are sometimes not resolvable
                assert (
                    move_request["pp"] == move.current_pp
                ), f"{move_request['pp']} != {move.current_pp}"
            if "maxpp" in move_request:
                assert (
                    move_request["maxpp"] == move.max_pp
                ), f"{move_request['maxpp']} != {move.max_pp}"
            assert move.target is not None
            if "target" in move_request:
                if move.non_ghost_target and PokemonType.GHOST not in self.types:
                    target_name = Target.SELF.name
                elif (
                    move.id == "terastarstorm"
                    and not self.fainted
                    and self.is_terastallized
                    and self.tera_type == PokemonType.STELLAR
                ):
                    target_name = Target.ALL_ADJACENT_FOES.name
                elif move.id == "pollenpuff" and Effect.HEAL_BLOCK in self.effects:
                    target_name = Target.ADJACENT_FOE.name
                else:
                    target_name = move.target.name
                assert (
                    Target.from_showdown_message(move_request["target"]).name
                    == target_name
                ), f"{Target.from_showdown_message(move_request['target']).name} != {target_name}"

    def clear_active(self):
        self._active = False

    def clear_boosts(self):
        for stat in self._boosts:
            self._boosts[stat] = 0

    def _clear_effects(self):
        for effect in [e for e in self._effects]:
            self.end_effect(effect.name)

    def clear_negative_boosts(self):
        for stat, value in self._boosts.items():
            if value < 0:
                self._boosts[stat] = 0

    def clear_positive_boosts(self):
        for stat, value in self._boosts.items():
            if value > 0:
                self._boosts[stat] = 0

    def copy_boosts(self, mon: Pokemon):
        self._boosts = dict(mon._boosts.items())

    def cure_status(self, status: Optional[str] = None):
        if status and Status[status.upper()] == self._status:
            self._status = None
            self._status_counter = 0
        elif status is None and not self.fainted:
            self._status = None

    def damage(self, hp_status: str):
        self.set_hp_status(hp_status)

    def end_effect(self, effect_str: str):
        effect = Effect.from_showdown_message(effect_str)
        if effect in self._effects:
            self._effects.pop(effect)

        if effect == Effect.TYPECHANGE:
            self._temporary_types = []

        elif effect == Effect.SKILL_SWAP:
            self._temporary_ability = None

        # Because of Showdown protocols, Protosynthesis and Quark Drive
        # both get two Effects added to the mon, but only have one protocol
        # message when the Effect ends
        elif effect == Effect.QUARK_DRIVE:
            for effect in [
                Effect.QUARKDRIVEATK,
                Effect.QUARKDRIVEDEF,
                Effect.QUARKDRIVESPA,
                Effect.QUARKDRIVESPD,
                Effect.QUARKDRIVESPE,
            ]:
                if effect in self._effects:
                    self._effects.pop(effect)

        elif effect == Effect.PROTOSYNTHESIS:
            for effect in [
                Effect.PROTOSYNTHESISATK,
                Effect.PROTOSYNTHESISDEF,
                Effect.PROTOSYNTHESISSPA,
                Effect.PROTOSYNTHESISSPD,
                Effect.PROTOSYNTHESISSPE,
            ]:
                if effect in self._effects:
                    self._effects.pop(effect)

    def end_item(self, item: str):
        self._item = None

        if item == "powerherb":
            self._preparing_move = None
            self._preparing_target = False

    def end_turn(self):
        self._active_turns += 1
        if self._status == Status.TOX:
            self._status_counter += 1
        for effect in list(self.effects.keys()):
            if effect.is_turn_countable:
                self.effects[effect] += 1
            if effect.ends_on_turn:
                self.end_effect(effect.name)

    def faint(self):
        self._current_hp = 0
        self._status = Status.FNT
        self.temporary_ability = None
        self._temporary_base_stats = None
        self._moves._transform_moves = None
        self._moves.mimic_move = None
        self._clear_effects()

    def forme_change(self, species: str):
        species = species.split(",")[0]
        self._update_from_pokedex(species, store_species=False)

    def identifies_as(self, ident: str) -> bool:
        return self.base_species == to_id_str(ident) or self.base_species in [
            to_id_str(substr) for substr in ident.split("-")
        ]

    def invert_boosts(self):
        self._boosts = {k: -v for k, v in self._boosts.items()}

    def mega_evolve(self, stone: str):
        species_id_str = to_id_str(self.species)
        mega_species = (
            species_id_str + "mega"
            if not species_id_str.endswith("mega")
            else species_id_str
        )
        self.temporary_ability = None
        if mega_species in GenData.from_gen(self.gen).pokedex:
            self._update_from_pokedex(mega_species, store_species=False)
        elif stone[-1] in "XYxy":
            mega_species = mega_species + stone[-1].lower()
            self._update_from_pokedex(mega_species, store_species=False)

    def moved(
        self,
        move_id: str,
        failed: bool = False,
        use: bool = True,
        reveal: bool = True,
        pressure: bool = False,
    ):
        self._must_recharge = False
        self._preparing_move = None
        self._preparing_target = None
        move = None
        if reveal:
            move = self._add_move(move_id)
        if use:
            if move is not None:
                move.use(pressure)
            for m in self.moves.values():
                m._is_last_used = m is move

        if move is not None and move.is_protect_counter and not failed:
            self._protect_counter += 1
        else:
            self._protect_counter = 0

        if self._status == Status.SLP:
            self._status_counter += 1

        # Handle silent effect ending
        if Effect.GLAIVE_RUSH in self.effects:
            self.end_effect("Glaive Rush")
        elif (
            Effect.CHARGE in self.effects
            and isinstance(move, Move)
            and move.base_power > 0
            and move.type == PokemonType.ELECTRIC
            and use
        ):
            self.end_effect("Charge")
        elif (
            Effect.FLASH_FIRE in self.effects
            and isinstance(move, Move)
            and move.base_power > 0
            and move.type == PokemonType.FIRE
            and use
        ):
            self.end_effect("Flash Fire")

    def prepare(self, move_id: str, target: Optional[Pokemon]):
        ident = Move.retrieve_id(move_id)
        if ident in self.moves:
            move = self.moves[ident]
        else:
            # could happen if, for example, move is copied from opponent via copycat
            move = Move(ident, raw_id=move_id, gen=self.gen)
        self._preparing_move = move
        self._preparing_target = target

    def primal(self):
        species_id_str = to_id_str(self._species)
        primal_species = (
            species_id_str + "primal"
            if not species_id_str.endswith("primal")
            else species_id_str
        )
        self._update_from_pokedex(primal_species, store_species=False)

    def set_boost(self, stat: str, amount: int):
        assert (
            abs(amount) <= 6
        ), f"{stat} of mon {self._species} is not <= 6. Got {amount}"
        self._boosts[stat] = int(amount)

    def set_hp(self, hp_status: str):
        self.set_hp_status(hp_status)

    # Param `store` dictates whether we should store the HP as a mon's stats
    def set_hp_status(self, hp_status: str, store=False):
        if hp_status == "0 fnt":
            self.faint()
            return

        if " " in hp_status:
            hp, status = hp_status.split(" ")
            self._status = Status[status.upper()]

            # Handle silent effect ending
            if self._status == Status.SLP and Effect.YAWN in self.effects:
                self.end_effect("yawn")
        else:
            hp = hp_status
            self._status = None

        current_hp, max_hp = "".join([c for c in hp if c in "0123456789/"]).split("/")
        self._current_hp = int(current_hp)
        self._max_hp = int(max_hp)

        if store:
            self._stats["hp"] = self._max_hp

    def start_effect(self, effect_str: str, details: Optional[Any] = None):
        effect = Effect.from_showdown_message(effect_str)
        if effect not in self._effects:
            self._effects[effect] = 0
        elif effect.is_action_countable:
            self._effects[effect] += 1

        if effect.breaks_protect:
            self._protect_counter = 0

        if effect == Effect.TYPECHANGE and details is not None:
            self._temporary_types = []
            for type_ in details.split("/"):
                self._temporary_types.append(PokemonType.from_name(type_))
        elif effect == Effect.SKILL_SWAP and details:
            # Skill Swap reveals a mon's ability
            if self.ability is None:
                self.ability = details[1]
            self.temporary_ability = details[0]

    def _swap_boosts(self):
        self._boosts["atk"], self._boosts["spa"] = (
            self._boosts["spa"],
            self._boosts["atk"],
        )

    def switch_in(self, details: Optional[str] = None):
        self._active = True
        if details:
            self._update_from_details(details)
        self._revealed = True

    def switch_out(self, fields: Dict[Field, int]):
        if (
            self.ability == "regenerator"
            and (
                self.item == "abilityshield"
                or Field.NEUTRALIZING_GAS not in fields.keys()
            )
            and self.status != Status.FNT
        ):
            self._current_hp = min(int(self.current_hp + self.max_hp / 3), self.max_hp)

        self._active = False
        self.clear_boosts()
        self._clear_effects()
        self._active_turns = 0
        self._must_recharge = False
        self._preparing_move = None
        self._preparing_target = None
        self._protect_counter = 0
        self.temporary_ability = None
        self._temporary_base_stats = None
        self._moves._transform_moves = None
        self._moves.mimic_move = None
        self._temporary_types = []

        if self._status == Status.TOX:
            self._status_counter = 0

        for move in self.moves.values():
            move._is_last_used = False

    def terastallize(self, type_: str):
        self._terastallized_type = PokemonType.from_name(type_)
        self._terastallized = True
        self._temporary_types = []

    def transform(self, into: Pokemon):
        dex_entry = GenData.from_gen(self.gen).pokedex[into.species]
        self._heightm = dex_entry["heightm"]
        self._weightkg = dex_entry["weightkg"]
        self._temporary_base_stats = dex_entry["baseStats"]
        if into.ability is not None:
            self._temporary_ability = into.ability
        self._temporary_types = [PokemonType.from_name(t) for t in dex_entry["types"]]
        self._moves._transform_moves = MoveSet(
            {m.id: Move(m.id, m.gen, from_transform=True) for m in into.moves.values()}
        )
        self._boosts = into.boosts.copy()

    def _update_from_pokedex(self, species: str, store_species: bool = True):
        species = to_id_str(species)
        dex_entry = GenData.from_gen(self.gen).pokedex[species]
        if store_species:
            self._species = species
        self._base_stats = dex_entry["baseStats"]

        self._type_1 = PokemonType.from_name(dex_entry["types"][0])
        if len(dex_entry["types"]) == 1:
            self._type_2 = None
        else:
            self._type_2 = PokemonType.from_name(dex_entry["types"][1])

        if "forme" in dex_entry and (
            dex_entry["forme"].startswith("Mega")
            or dex_entry["forme"] in ["Primal", "Stellar", "Terastal"]
            or dex_entry["forme"].endswith("-Tera")
        ):
            self.forme_change_ability = dex_entry["abilities"]["0"]
        elif self.forme_change_ability is None:
            self._possible_abilities = [
                to_id_str(ability) for ability in dex_entry["abilities"].values()
            ]
            if len(self._possible_abilities) == 1 and self.gen >= 3:
                self._ability = self._possible_abilities[0]
        else:
            self.forme_change_ability = None

        self._heightm = dex_entry["heightm"]
        self._weightkg = dex_entry["weightkg"]

    def _update_from_details(self, details: str):
        if details == self._last_details:
            return
        else:
            self._last_details = details

        if ", shiny" in details:
            self._shiny = True
            details = details.replace(", shiny", "")
        else:
            self._shiny = False

        split_details = details.split(", ")

        gender = None
        level = None

        for split_detail in split_details:
            if split_detail.startswith("tera:"):
                self._terastallized_type = PokemonType.from_name(split_detail[5:])

                split_details.remove(split_detail)
                break

        if len(split_details) == 3:
            species, level, gender = split_details
        elif len(split_details) == 2:
            if split_details[1].startswith("L"):
                species, level = split_details
            else:
                species, gender = split_details
        else:
            species = to_id_str(split_details[0])

        if gender:
            self._gender = PokemonGender.from_request_details(gender)
        else:
            self._gender = PokemonGender.NEUTRAL

        if level:
            self._level = int(level[1:])
        else:
            self._level = 100

        if species != self._species:
            self._update_from_pokedex(species)

    def update_from_request(self, request_pokemon: Dict[str, Any]):
        self._active = request_pokemon["active"]

        if self.ability is None:
            self.ability = request_pokemon["baseAbility"]
        if (
            "ability" in request_pokemon
            and request_pokemon["ability"] != request_pokemon["baseAbility"]
        ):
            self.temporary_ability = request_pokemon["ability"]

        self._last_request = request_pokemon

        condition = request_pokemon["condition"]
        self.set_hp_status(condition, store=True)

        self._name = request_pokemon["ident"][4:]
        self._item = request_pokemon["item"]

        details = request_pokemon["details"]
        self._update_from_details(details)

        for move in request_pokemon["moves"]:
            self._add_move(move)

        if "stats" in request_pokemon:
            for stat in request_pokemon["stats"]:
                self._stats[stat] = request_pokemon["stats"][stat]

    def _update_from_teambuilder(self, tb: TeambuilderPokemon):
        if tb.nickname is not None and tb.species is None:
            self._update_from_pokedex(tb.nickname)
        elif tb.nickname is not None and tb.species is not None:
            self._update_from_pokedex(tb.species)
            self._name = tb.nickname
        elif tb.nickname is None and tb.species is not None:
            self._update_from_pokedex(tb.species)
        else:
            raise ValueError(
                "TeambuilderPokemon must have either a nickname or species", tb
            )

        if tb.level:
            self._level = tb.level
        self._ability = to_id_str(tb.ability) if tb.ability is not None else None
        self._item = to_id_str(tb.item) if tb.item else None
        if tb.gender:
            self._gender = PokemonGender.from_request_details(tb.gender)
        self._shiny = tb.shiny
        if tb.tera_type:
            self._terastallized_type = PokemonType.from_name(tb.tera_type)
        self._moves = MoveSet({})
        for move_str in tb.moves:
            move = Move(Move.retrieve_id(move_str), gen=self.gen)
            self._moves[move.id] = move

        if not all(e == 0 for e in tb.evs):
            self._evs = tb.evs
            self._ivs = tb.ivs
            self._nature = tb.nature.lower() if tb.nature is not None else "serious"

        if tb.level:
            nature = tb.nature.lower() if tb.nature else "serious"
            self._stats = {}
            stats = compute_raw_stats(
                self._species,
                tb.evs,
                tb.ivs,
                tb.level,
                nature,
                GenData.from_gen(self.gen),
            )
            for stat, val in zip(["hp", "atk", "def", "spa", "spd", "spe"], stats):
                self._stats[stat] = val

    def was_illusioned(self, fields: Dict[Field, int]):
        self._current_hp = None
        self._max_hp = None
        self._status = None

        last_request = self._last_request
        self._last_request = None

        if last_request:
            self.update_from_request(last_request)

        self.switch_out(fields)

    def available_moves_from_request(self, request: Dict[str, Any]) -> List[Move]:
        moves: List[Move] = []
        if Effect.COMMANDER in self.effects:
            return []

        request_moves: List[str] = [
            move["id"] for move in request["moves"] if not move.get("disabled", False)
        ]
        for move in request_moves:
            if move in self.moves:
                if self.is_dynamaxed:
                    moves.append(self.moves[move].dynamaxed)
                else:
                    moves.append(self.moves[move])
            elif move in SPECIAL_MOVES:
                moves.append(Move(move, gen=self.gen))
            elif (
                move == "hiddenpower"
                and len([m for m in self.moves if m.startswith("hiddenpower")]) == 1
            ):
                moves.append(
                    [v for m, v in self.moves.items() if m.startswith("hiddenpower")][0]
                )
            else:
                assert self.ability == "dancer" or {
                    "copycat",
                    "metronome",
                    "mefirst",
                    "mirrormove",
                    "assist",
                }.intersection(self.moves), (
                    f"Error with move {move}. Expected self.moves to contain copycat, "
                    "metronome, mefirst, mirrormove, or assist, or to have the ability "
                    f"dancer. Got {self.moves}, ability: {self.ability}"
                )
                moves.append(Move(move, gen=self.gen))
        return moves

    def damage_multiplier(self, type_or_move: Union[PokemonType, Move]) -> float:
        """
        Returns the damage multiplier associated with a given type or move on this
        pokemon.

        This method is a shortcut for PokemonType.damage_multiplier with relevant types.

        :param type_or_move: The type or move of interest.
        :type type_or_move: PokemonType or Move
        :return: The damage multiplier associated with given type on the pokemon.
        :rtype: float
        """
        if isinstance(type_or_move, Move):
            type_or_move = type_or_move.type
        return type_or_move.damage_multiplier(
            self.type_1, self.type_2, type_chart=GenData.from_gen(self.gen).type_chart
        )

    @property
    def ability(self) -> Optional[str]:
        """
        :return: The pokemon's ability. None if unknown or removed.
        :rtype: str, optional
        """
        if self.temporary_ability is not None:
            return self.temporary_ability
        elif self.forme_change_ability is not None:
            return self.forme_change_ability
        else:
            return self._ability

    @ability.setter
    def ability(self, ability: str):
        if self._ability is None:
            self._ability = to_id_str(ability)
        else:
            self._temporary_ability = to_id_str(ability)

    @property
    def active(self) -> Optional[bool]:
        """
        :return: Boolean indicating whether the pokemon is active.
        :rtype: bool
        """
        return self._active

    @active.setter
    def active(self, value: Optional[bool]):
        self.active = value

    @property
    def available_z_moves(self) -> List[Move]:
        """
        Caution: this property is not properly tested yet.

        :return: The set of moves that pokemon can use as z-moves.
        :rtype: List[Move]
        """
        if isinstance(self.item, str) and self.item.endswith("iumz"):
            type_, move = Z_CRYSTAL[self.item]
            if type_:
                return [
                    move
                    for move in self.moves.values()
                    if move.type == type_ and move.can_z_move
                ]
            elif move in self.moves:
                return [self.moves[move]]
        return []

    @property
    def base_ability(self) -> Optional[str]:
        """
        :return: The pokemon's base ability. None if unknown.
        :rtype: str, optional
        """
        return self._forme_change_ability or self._ability

    @property
    def base_moves(self) -> Dict[str, Move]:
        """
        :return: The pokemon's underlying move dictionary. When transformed, this
            returns the temporary move set; otherwise it returns the learned moves.
        :rtype: Dict[str, Move]
        """
        return self._moves.base_moves

    @property
    def base_species(self) -> str:
        """
        :return: The pokemon's base species.
        :rtype: str
        """
        dex_entry = GenData.from_gen(self.gen).pokedex[self._species]
        return to_id_str(dex_entry["baseSpecies"])

    @property
    def base_stats(self) -> Dict[str, int]:
        """
        :return: The pokemon's base stats.
        :rtype: Dict[str, int]
        """
        return (
            self._temporary_base_stats
            if self._temporary_base_stats is not None
            else self._base_stats
        )

    @property
    def base_types(self) -> List[PokemonType]:
        """
        :return: The pokemon's non-Tera types, accounting for temporary type changes.
            Unlike `types`, this ignores Terastallization.
        :rtype: List[PokemonType]
        """
        if len(self._temporary_types) > 0:
            return self._temporary_types
        else:
            types = [self._type_1]
            if self._type_2 is not None:
                types.append(self._type_2)
            return types

    @property
    def boosts(self) -> Dict[str, int]:
        """
        :return: The pokemon's boosts.
        :rtype: Dict[str, int]
        """
        return self._boosts

    @boosts.setter
    def boosts(self, value: Dict[str, int]):
        self._boosts = value

    @property
    def current_hp(self) -> int:
        """
        :return: The pokemon's current hp. For your pokemons, this is the actual value.
            For opponent's pokemon, this value depends on showdown information: it can
            be on a scale from 0 to 100 or on a pixel scale.
        :rtype: int
        """
        return self._current_hp or 0

    @property
    def current_hp_fraction(self) -> float:
        """
        :return: The pokemon's current remaining hp fraction.
        :rtype: float
        """
        if self.current_hp:
            return self.current_hp / self.max_hp
        return 0

    @property
    def effects(self) -> Dict[Effect, int]:
        """
        :return: A Dict mapping the effects currently affecting the pokemon and the
            associated counter.
        :rtype: Dict[Effect, int]
        """
        return self._effects

    @property
    def evs(self) -> list[int] | None:
        """
        :return: The pokemon's EVs as a list of [HP, Atk, Def, SpA, SpD, Spe].
            None if unknown (e.g. opponent's pokemon).
        :rtype: List[int] | None
        """
        return self._evs

    @property
    def fainted(self) -> bool:
        """
        :return: Whether the pokemon has fainted.
        :rtype: bool
        """
        return Status.FNT == self._status

    @property
    def first_turn(self) -> bool:
        """
        :return: Whether this is this pokemon's first action since its last switch in.
        :rtype: bool
        """
        return self._active_turns == 1

    @property
    def forme_change_ability(self) -> Optional[str]:
        """
        :return: The pokemon's ability after changing forme. None if the pokemon hasn't changed forme.
        :rtype: str, optional
        """
        return self._forme_change_ability

    @forme_change_ability.setter
    def forme_change_ability(self, ability: Optional[str]):
        self._forme_change_ability = to_id_str(ability) if ability is not None else None

    @property
    def gen(self) -> int:
        """
        :return: The generation of the pokemon.
        :rtype: int
        """
        return self._gen

    @property
    def gender(self) -> Optional[PokemonGender]:
        """
        :return: The pokemon's gender.
        :rtype: PokemonGender, optional
        """
        return self._gender

    @property
    def height(self) -> float:
        """
        :return: The pokemon's height, in meters.
        :rtype: float
        """
        return self._heightm

    @property
    def hp_status(self) -> str:
        if self.status == Status.FNT:
            s = "0 fnt"
        else:
            s = f"{self.current_hp}/{self.max_hp}"
            if self.status is not None:
                s += f" {self.status.name.lower()}"
        return s

    def identifier(self, player_role: str) -> str:
        """ "
        :param player_role: The player's role in the battle (p1 or p2)
        :type player_role: str
        :return: The pokemon's identifier, which can be used to identify it in Showdown logs
        """
        assert player_role in ["p1", "p2"]
        return player_role + ": " + self.name

    @property
    def is_dynamaxed(self) -> bool:
        """
        :return: Whether the pokemon is currently dynamaxed
        :rtype: bool
        """
        return Effect.DYNAMAX in self.effects

    @property
    def is_terastallized(self) -> bool:
        """
        :return: Whether the pokemon is currently terastallized
        :rtype: bool
        """
        return self._terastallized

    @property
    def item(self) -> Optional[str]:
        """
        :return: The pokemon's item.
        :rtype: str | None
        """
        return self._item

    @item.setter
    def item(self, item: Optional[str]):
        self._item = to_id_str(item) if item is not None else None

    @property
    def ivs(self) -> list[int] | None:
        """
        :return: The pokemon's IVs as a list of [HP, Atk, Def, SpA, SpD, Spe].
            None if unknown (e.g. opponent's pokemon).
        :rtype: List[int] | None
        """
        return self._ivs

    @property
    def last_move(self) -> Optional[Move]:
        """
        :return: The last move used by this pokemon, or None.
        :rtype: Move | None
        """
        for move in self.moves.values():
            if move.is_last_used:
                return move
        return None

    @property
    def level(self) -> int:
        """
        :return: The pokemon's level.
        :rtype: int
        """
        return self._level

    @property
    def max_hp(self) -> int:
        """
        :return: The pokemon's max hp. For your pokemons, this is the actual value.
            For opponent's pokemon, this value depends on showdown information: it can
            be on a scale from 0 to 100 or on a pixel scale.
        :rtype: int
        """
        return self._max_hp or 0

    @property
    def mimic_move(self) -> Move | None:
        """
        :return: The move currently copied by Mimic, if any.
        :rtype: Move | None
        """
        return self._moves.mimic_move

    @property
    def moves(self) -> Dict[str, Move]:
        """
        :return: A dictionary of the pokemon's known moves.
        :rtype: Dict[str, Move]
        """
        return self._moves.moves

    @property
    def must_recharge(self) -> bool:
        """
        :return: A boolean indicating whether the pokemon must recharge.
        :rtype: bool
        """
        return self._must_recharge

    @must_recharge.setter
    def must_recharge(self, value: bool):
        self._must_recharge = value

    @property
    def name(self) -> str:
        """
        :return: The pokemon's name, which can be used to create Showdown's
            identifier.
        :rtype: str
        """
        if self._name is not None:
            return self._name
        else:
            dex_entry = GenData.from_gen(self.gen).pokedex[self._species]
            if dex_entry["baseSpecies"].islower():
                return dex_entry["name"]
            else:
                return dex_entry["baseSpecies"]

    @property
    def nature(self) -> str | None:
        """
        :return: The pokemon's nature if known. None if not provided by the teambuilder.
        :rtype: str | None
        """
        return self._nature

    @property
    def pokeball(self) -> Optional[str]:
        """
        :return: The pokeball in which is the pokemon.
        :rtype: str | None
        """
        if self._last_request is not None:
            return self._last_request.get("pokeball", None)
        return None

    @property
    def possible_abilities(self) -> List[str]:
        """
        :return: The list of possible abilities for this pokemon.
        :rtype: List[str]
        """
        return self._possible_abilities

    @property
    def preparing(self) -> bool:
        """
        :return: Whether this pokemon is preparing a multi-turn move.
        :rtype: bool
        """
        return bool(self._preparing_target) or bool(self._preparing_move)

    @property
    def preparing_target(self) -> Optional[Union[bool, Pokemon]]:
        """
        :return: The moves target - optional.
        :rtype: Any
        """
        return self._preparing_target

    @property
    def preparing_move(self) -> Optional[Move]:
        """
        :return: The move being prepared - optional.
        :rtype: Move, optional
        """
        return self._preparing_move

    @property
    def protect_counter(self) -> int:
        """
        :return: How many protect-like moves where used in a row by this pokemon.
        :rtype: int
        """
        return self._protect_counter

    @property
    def revealed(self) -> bool:
        """
        :return: Whether this pokemon has appeared in the current battle.
        :rtype: bool
        """
        return self._revealed

    @property
    def shiny(self) -> bool:
        """
        :return: Whether this pokemon is shiny.
        :rtype: bool
        """
        return bool(self._shiny)

    @property
    def species(self) -> str:
        """
        :return: The pokemon's species.
        :rtype: str | None
        """
        return self._species

    @property
    def stats(self) -> Dict[str, Optional[int]]:
        """
        :return: The pokemon's stats, as a dictionary.
        :rtype: Dict[str, int | None]
        """
        return self._stats

    @stats.setter
    def stats(self, stats: Dict[str, Optional[int]]):
        self._stats = stats

    @property
    def status(self) -> Optional[Status]:
        """
        :return: The pokemon's status.
        :rtype: Optional[Status]
        """
        return self._status

    @status.setter
    def status(self, status: Optional[Union[Status, str]]):
        self._status = Status[status.upper()] if isinstance(status, str) else status

    @property
    def status_counter(self) -> int:
        """
        :return: The pokemon's status turn count. Only counts TOXIC and SLEEP statuses.
        :rtype: int
        """
        return self._status_counter

    @property
    def stab_multiplier(self) -> float:
        """
        :return: The pokemon's STAB multiplier.
        :rtype: float
        """
        # https://bulbapedia.bulbagarden.net/wiki/Terastal_phenomenon#Effects
        if (
            self._terastallized
            and (
                self.tera_type in (self._type_1, self._type_2)
                or self.tera_type in self._temporary_types
            )
            and self.ability == "adaptability"
        ):
            return 2.25
        elif self._terastallized and (
            self.tera_type in (self._type_1, self._type_2)
            or self.tera_type in self._temporary_types
        ):
            return 2
        elif self.ability == "adaptability":
            return 2
        return 1.5

    @property
    def selected_in_teampreview(self) -> bool:
        """
        :return: Whether this pokemon was selected in teampreview.
        :rtype: bool
        """
        return self._selected_in_teampreview

    @property
    def temporary_ability(self) -> str | None:
        """
        :return: The pokemon's temporary ability, None if none is set.
        :rtype: Optional[str]
        """
        return self._temporary_ability

    @temporary_ability.setter
    def temporary_ability(self, ability: str | None):
        self._temporary_ability = to_id_str(ability) if ability is not None else None

    @property
    def tera_type(self) -> Optional[PokemonType]:
        """
        :return: The Tera Type of the Pokemon, None if unnown
        :rtype: Optional[PokemonType]
        """
        return self._terastallized_type

    @property
    def transformed(self) -> bool:
        """
        :return: Whether the pokemon is currently transformed.
        :rtype: bool
        """
        return self._moves._transform_moves is not None

    @property
    def type_1(self) -> PokemonType:
        """
        :return: The pokemon's first type.
        :rtype: PokemonType
        """
        if self._terastallized and self._terastallized_type is not None:
            return self._terastallized_type
        elif len(self._temporary_types) > 0:
            return self._temporary_types[0]
        return self._type_1

    @property
    def type_2(self) -> Optional[PokemonType]:
        """
        :return: The pokemon's second type.
        :rtype: Optional[PokemonType]
        """
        if self._terastallized or len(self._temporary_types) == 1:
            return None
        elif len(self._temporary_types) > 1:
            return self._temporary_types[1]
        return self._type_2

    @property
    def types(self) -> List[PokemonType]:
        """
        :return: The pokemon's types. Not a Tuple because moves can add types
            and also remove types, and override them too.
        :rtype: Tuple[PokemonType]
        """
        if len(self._temporary_types) > 0:
            return self._temporary_types
        else:
            types = [self.type_1]
            if self.type_2 is not None:
                types.append(self.type_2)
            return types

    @property
    def weight(self) -> float:
        """
        :return: The pokemon's weight, in kilograms.
        :rtype: float
        """
        return self._weightkg
